<?php

namespace ProgrammaBill;

use Symfony\Component\HttpKernel\Exception\AccessDeniedHttpException;
use Symfony\Component\HttpKernel\Event\ControllerEvent;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpKernel\KernelEvents;
use Nyholm\Psr7\Factory\Psr17Factory;
use Symfony\Bridge\PsrHttpMessage\Factory\PsrHttpFactory;
use Symfony\Component\HttpFoundation\Request;
use CloudObjects\SDK\ObjectRetriever;
use CloudObjects\SDK\AccountGateway\AccountContext;
use CloudObjects\SDK\Helpers\SharedSecretAuthentication;
use Doctrine\ORM\EntityManagerInterface;
use ProgrammaBill\Entities\Actor,
    ProgrammaBill\Entities\Marketplace;

class AuthenticationSubscriber implements EventSubscriberInterface {

    private $actorRepository;
    private $marketplaceRepository;
    private $sharedSecretAuthentication;
    private $actor;
    private $marketplace;
    private $objectRetriever;

    public function __construct(EntityManagerInterface $entityManager, ObjectRetriever $objectRetriever,
            SharedSecretAuthentication $sharedSecretAuthentication) {
        
        $this->actorRepository = $entityManager->getRepository(Actor::class);
        $this->marketplaceRepository = $entityManager->getRepository(Marketplace::class);
        $this->sharedSecretAuthentication = $sharedSecretAuthentication;
        $this->objectRetriever = $objectRetriever;
    }

    private function convertRequest(Request $request) {
        $psr17Factory = new Psr17Factory;
        $psrHttpFactory = new PsrHttpFactory($psr17Factory, $psr17Factory, $psr17Factory, $psr17Factory);
        return $psrHttpFactory->createRequest($request);
    }
    
    public function onKernelController(ControllerEvent $event) {
        $request = $event->getRequest();
        if ($request->getUser() != null && $request->getPassword() != null) {
            // HTTP Basic-authenticated request -> Service
            if ($this->sharedSecretAuthentication->verify($request->getUser(),
                    $request->getPassword()) != SharedSecretAuthentication::RESULT_OK)
                throw new AccessDeniedHttpException;
            
            $this->marketplace = $this->marketplaceRepository->fetchOrCreateWithNamespace($request->getUser());
            $this->marketplace->setObjectRetriever($this->objectRetriever);
        } else {
            // Authenticated request coming from Account Gateway -> Actor            
            $context = AccountContext::fromPsrRequest($this->convertRequest($event->getRequest()));
            if (!isset($context))
                throw new AccessDeniedHttpException;

            $this->actor = $this->actorRepository->fetchOrCreateFor($context);
            $this->actor->getMarketplace()->setObjectRetriever($this->objectRetriever);
        }
    }

    public static function getSubscribedEvents() {
        return [
            KernelEvents::CONTROLLER => 'onKernelController',
        ];
    }

	public function getActor() {
        return $this->actor;
    }

    public function getMarketplace() {
        return $this->marketplace;
    }

}