<?php

namespace ProgrammaBill\Commands\Cron;

use Exception;
use Symfony\Component\Console\Command\Command;
use Symfony\Component\Console\Input\InputInterface,
    Symfony\Component\Console\Input\InputArgument;
use Symfony\Component\Console\Output\OutputInterface;
use Doctrine\ORM\EntityManagerInterface;
use Carbon\Carbon;
use GuzzleHttp\Client;
use ProgrammaBill\Entities\Marketplace,
    ProgrammaBill\Entities\Invoice,
    ProgrammaBill\Entities\InvoiceAction;

class NotifyOverdueInvoices extends Command {

    private $em;

    protected static $defaultName = 'app-cron:notify-overdue-invoices';
    protected static $defaultDescription = 'Finds all overdue invoices and sends a reminder (if configured).';

    public function __construct(EntityManagerInterface $em) {
        $this->em = $em;
        parent::__construct();
    }

    protected function execute(InputInterface $input, OutputInterface $output): int {
        $marketplaces = $this->em->getRepository(Marketplace::class)->findAll();
        $invoiceRepository = $this->em->getRepository(Invoice::class);
        $client = new Client;
        $actionType_1 = InvoiceAction::actionTypeNativeToString(InvoiceAction::TYPE_NOTIFIED_OVERDUE);
        $actionType_2 = InvoiceAction::actionTypeNativeToString(InvoiceAction::TYPE_NOTIFIED_IN_DEFAULT);

        foreach ($marketplaces as $marketplace) {
            // Determine configuration for marketplace
            // TODO: replace hardcoded choice with condition check
            if (in_array($marketplace->getServiceDomain(), [ 'welmoa.de' ])) {
                $webhookUrl = 'https://menoia.de/webhooks/programmabill'; // TODO: get from marketplace configuration
            } else {
                // nothing to do for this marketplace
                continue;
            }

            // Get invoices
            $invoices = $invoiceRepository->getAllOverdueInvoices($marketplace);

            // Filter to invoices that need to receive the notification
            $invoicesToNotify = $invoices->filter(function($invoice) use ($marketplace) {
                
                // Find previous "overdue" notifications
                $notifications = $invoice->findActionsWithType(InvoiceAction::TYPE_NOTIFIED_OVERDUE);

                if (count($notifications) == 0) {
                    // Never notified, so we definitely need to send one
                    return true;
                }

                if (count($notifications) == 1) {
                    // Once notified, so should go into default after delay
                    $inDefault = (Carbon::now()->subDays($marketplace->getInDefaultAfter()) > $notifications[0]->getCreatedTimestamp());
                    if ($inDefault) {
                        // Check whether we already notified about default
                        $notifications = $invoice->findActionsWithType(InvoiceAction::TYPE_NOTIFIED_IN_DEFAULT);
                        
                        return (count($notifications) == 0);
                    }
                }

                return false;
            });

            // Send "overdue" notifications and mark notified on success
            foreach ($invoicesToNotify as $invoice) {
                $request = [
                    'type' => 'invoice_action',
                    'action_type' => count($invoice->findActionsWithType(InvoiceAction::TYPE_NOTIFIED_OVERDUE)) == 0
                        ? $actionType_1 // first notification = "overdue"
                        : $actionType_2, // second notification = "in default"
                    'transaction_id' => $invoice->getTransaction()->getId(),
                    'invoice_type' => $invoice->getInvoiceTypeAsString(),
                    'invoice_number' => $invoice->getNumber()
                ];

                try {                    
                    // Send webhook
                    $client->post($webhookUrl, [
                        'form_params' => $request
                    ]);

                    // After sending webhook, add action
                    $invoice->getActions()->add(
                        new InvoiceAction($invoice, InvoiceAction::TYPE_NOTIFIED_OVERDUE)
                    );

                    echo "successfully triggered -> " . json_encode($request) . "\n\n";
                } catch (Exception $e) {
                    // skip when webhook failed
                    echo "failed -> " . json_encode($request) . "\n\n";
                }
            }

            // Persist actions to database
            $this->em->flush();
        }

        return 0;
    }

}
