<?php

namespace ProgrammaBill\Entities;

use Doctrine\ORM\Mapping AS ORM;
use Doctrine\Common\Collections\ArrayCollection;
use Carbon\Carbon;
use Webmozart\Assert\Assert;
use Punic\Territory;
use CloudObjects\SDK\NodeReader;
use CloudObjects\SDK\AccountGateway\AAUIDParser,
    CloudObjects\SDK\AccountGateway\AccountContext;
use ProgrammaBill\Interfaces\APIObject,
    ProgrammaBill\Interfaces\Address;

/**
 * An Actor is an entity which can have transactions on a marketplace as both
 * merchant and customer; it can either be an individual user account
 * or a group account such as an organization.
 *
 * @ORM\Entity(repositoryClass="ProgrammaBill\Repositories\ActorRepository")
 * @ORM\Table(name="mi_actors", uniqueConstraints={
 *  @ORM\UniqueConstraint(name="email_marketplace_idx", columns={"email", "marketplace_id"})
 * })
 */
class Actor extends BaseEntity implements APIObject, Address, \ArrayAccess {

    const TYPE_UNKNOWN = 0;
    const TYPE_INDIVIDUAL = 1;
    const TYPE_GROUP = 2;

    /** @ORM\Column(type="string", length=16, unique=true) **/
    private $aauid;

    /** @ORM\Column(type="string", length=250) **/
    private $access_token;

    /** @ORM\Column(type="string", length=50) **/
    private $email;

    /** @ORM\Column(type="string", length=50, nullable=true) **/
    private $ext_user_id;

    /** @ORM\ManyToOne(targetEntity="Marketplace") **/
    private $marketplace;

    /** @ORM\Column(type="smallint") **/
    private $actor_type;

    /** @ORM\Column(type="string", length=70, nullable=true) **/
    private $org_name;

    /** @ORM\Column(type="string", length=70, nullable=true) **/
    private $dep_name;

    /** @ORM\Column(type="string", length=70, nullable=true) **/
    private $full_name;

    /** @ORM\Column(type="string", length=70, nullable=true) **/
    private $street_address;

    /** @ORM\Column(type="string", length=10, nullable=true) **/
    private $postal_code;

    /** @ORM\Column(type="string", length=70, nullable=true) **/
    private $location;

    /** @ORM\Column(type="string", length=2, nullable=true) **/
    private $country;

    /** @ORM\Column(type="string", length=50, nullable=true) **/
    private $eu_vat_id;

    /** @ORM\Column(type="boolean") */
    private $vat_exempt = false;

    /** @ORM\Column(type="string", length=50, nullable=true) **/
    private $company_tax_id;

    /** @ORM\Column(type="string", length=20, nullable=true) **/
    private $cost_center_id;

    /** @ORM\OneToMany(targetEntity="ActorDetail", mappedBy="actor", cascade={"persist"}) **/
    private $details;

    /** @ORM\ManyToOne(targetEntity="Identity") **/
    private $identity;

    /**
     * @ORM\OneToMany(targetEntity="Transaction", mappedBy="merchant", cascade={"persist"})
     * @ORM\OrderBy({"id" = "DESC"})
     */
    private $transactions_as_merchant;

    /**
     * @ORM\OneToMany(targetEntity="Transaction", mappedBy="customer", cascade={"persist"})
     * @ORM\OrderBy({"id" = "DESC"})
     */
    private $transactions_as_customer;

    // internal fields

    private $reader = null;
    private $context = null;

    private function getReader() : NodeReader {
        if (!isset($this->reader))
            $this->reader = new NodeReader([
                'prefixes' => [
                    'mi' => 'coid://marketplace-invoicing.cloudobjects.io/'
                ]
            ]);
        
        return $this->reader;
    }

    /**
     * Creates a new Actor.
     * @param Marketplace $marketplace Marketplace on which this actor acts
     * @param string $aauid CloudObjects account identifier
     * @param integer $actor_type Type of the actor, one of TYPE_UNKNOWN, TYPE_INDIVIDUAL, TYPE_GROUP
     */
    public function __construct(Marketplace $marketplace, string $aauid, int $actor_type) {
        Assert::length($aauid, 16);
        Assert::oneOf($actor_type, [ self::TYPE_UNKNOWN, self::TYPE_INDIVIDUAL, self::TYPE_GROUP ]);

        parent::__construct();
        $this->marketplace = $marketplace;
        $this->aauid = $aauid;
        $this->actor_type = $actor_type;

        $this->details = new ArrayCollection;
        $this->transactions_as_merchant = new ArrayCollection;
        $this->transactions_as_customer = new ArrayCollection;
    }

    /**
     * Get the CloudObjects account identifier for this actor.
     * @return string
     */
    public function getAAUID() : string {
        return $this->aauid;
    }

    /**
     * Specify an access token to access the Account Gateway
     * for this actor.
     * @param string|null $access_token
     * @return Actor self-reference for chaining
     */
    public function setAccessToken(?string $access_token) : Actor {
        $this->access_token = $access_token;
        return $this;
    }

    /**
     * Get the email address that uniquely identifies this actor
     * within the confines of the marketplace.
     * @return string
     */ 
    public function getEmail() : ?string {
        return $this->email;
    }

    /**
     * Set the email address that uniquely identifies this actor
     * within the confines of the marketplace.
     * @param string|null $email
     * @return Actor self-reference for chaining
     */ 
    public function setEmail(?string $email) : Actor {
        Assert::nullOrEmail($email);
        $this->email = $email;
        return $this;
    }

    /**
     * Get the external user ID set by the marketplace.
     * @return string|null
     */ 
    public function getExternalUserID() : ?string {
        return $this->ext_user_id;
    }

    /**
     * Set the external user ID used by the marketplace.
     * @param string|null $ext_user_id
     * @return Actor self-reference for chaining
     */ 
    public function setExternalUserID($ext_user_id) : Actor {
        $this->ext_user_id = $ext_user_id;
        return $this;
    }

    /**
     * Get the marketplace on which this actor acts.
     * @return Marketplace
     */
    public function getMarketplace() : Marketplace {
        return $this->marketplace;
    }

    /**
     * Get the actor's type.
     * @return integer one of TYPE_UNKNOWN, TYPE_INDIVIDUAL, TYPE_GROUP
     */
    public function getActorType() : int {
        return $this->actor_type;
    }

    /**
     * Get the organization name / common display name for this actor.
     * @return string|null
     */
    public function getOrgName() : ?string {
        return $this->org_name;
    }

    /**
     * Set the organization name / common display name for this actor.
     * @param string|null $org_name
     * @return Actor self-reference for chaining
     */
    public function setOrgName(?string $org_name) : Actor {
        $this->org_name = $org_name;
        return $this;
    }

    /**
     * Get the department name / subtitle display name for this actor.
     * @return string|null
     */
    public function getDepName() : ?string {
        return $this->dep_name;
    }

    /**
     * Set the department name / subtitle display name for this actor.
     * @param string|null $dep_name
     * @return Actor self-reference for chaining
     */
    public function setDepName(?string $dep_name) : Actor {
        $this->dep_name = $dep_name;
        return $this;
    }

    /**
     * Get the full name for this actor.
     * @return string|null
     */
    public function getFullName() : ?string {
        return $this->full_name;
    }

    /**
     * Set the full name for this actor.
     * @param string $full_name
     * @return Actor self-reference for chaining
     */
    public function setFullName(string $full_name) : Actor {
        $this->full_name = $full_name;
        return $this;
    }

    /**
     * Get the street address, typically a street name
     * and house number, for this actor.
     * @return string|null
     */
    public function getStreetAddress() : ?string {
        return $this->street_address;
    }

    /**
     * Set the street address, typically a street name
     * and house number, for this actor.
     * @param string|null $street_address
     * @return Actor self-reference for chaining
     */
    public function setStreetAddress(?string $street_address) : Actor {
        $this->street_address = $street_address;
        return $this;
    }

    /**
     * Get the postal code for the location of this actor.
     * @return string|null
     */
    public function getPostalCode() : ?string {
        return $this->postal_code;
    }

    /**
     * Set the postal code for the location of this actor.
     * @param string|null $postal_code
     * @return Actor self-reference for chaining
     */
    public function setPostalCode(?string $postal_code) : Actor {
        $this->postal_code = $postal_code;
        return $this;
    }

    /**
     * Get the name of the location, e.g. city or village,
     * of this actor.
     * @return string|null
     */
    public function getLocation() : ?string {
        return $this->location;
    }

    /**
     * Set the name of the location, e.g. city or village,
     * of this actor.
     * @param string|null $location
     * @return Actor self-reference for chaining
     */
    public function setLocation(?string $location) : Actor {
        $this->location = $location;
        return $this;
    }

    /**
     * Get the country in which the actor is located
     * as an ISO-3166-1 code.
     * @return string|null
     */
    public function getCountry() : ?string {
        return $this->country;
    }

    /**
     * Indicates whether the actor is located in the European Union.
     * @return bool
     */
    public function isLocatedInEU() : bool {
        return in_array($this->getCountry(), Territory::getChildTerritoryCodes('EU'));
    }

    /**
     * Set the country in which the actor is located
     * as an ISO-3166-1 code.
     * @param string $country
     * @return Actor self-reference for chaining
     */
    public function setCountry(string $country) : Actor {
        Assert::length($country, 2);
        $this->country = $country;
        return $this;
    }

    /**
     * Get the European VAT identification number for this actor.
     */
    public function getEUVatID() {
        return $this->eu_vat_id;
    }

    /**
     * Set the European VAT identification number for this actor.
     * @param string $eu_vat_id
     * @return Actor self-reference for chaining
     */
    public function setEUVatID(string $eu_vat_id) : Actor {
        $this->eu_vat_id = $eu_vat_id;
        return $this;
    }

    /**
     * Indicates whether this actor is exempt from charging VAT.
     */ 
    public function isVATExempt() : bool {
        return $this->vat_exempt;
    }

    /**
     * Specifies whether this actor is exempt from charging VAT.
     * @param bool $vat_exempt
     * @return Actor self-reference for chaining
     */ 
    public function setVATExempt(bool $vat_exempt) : Actor {
        $this->vat_exempt = $vat_exempt;
        return $this;
    }

    /**
     * Get the company Tax ID for this actor.
     * @return string|null
     */ 
    public function getCompanyTaxID() : ?string {
        return $this->company_tax_id;
    }

    /**
     * Set the company Tax ID for this actor.
     * @param string $company_tax_id
     * @return Actor self-reference for chaining
     */ 
    public function setCompanyTaxID(string $company_tax_id) : Actor {
        $this->company_tax_id = $company_tax_id;
        return $this;
    }

    /**
     * Get the cost center ID for this actor.
     * @return string|null
     */ 
    public function getCostCenterID() : ?string {
        return $this->cost_center_id;
    }

    /**
     * Set the cost center ID for this actor.
     * @param string $cost_center_id
     * @return Actor self-reference for chaining
     */ 
    public function setCostCenterID(string $cost_center_id) : Actor {
        $this->cost_center_id = $cost_center_id;
        return $this;
    }

    /**
     * Get the marketplace transactions in which this actor was the merchant.
     * @return ArrayCollection
     */
    public function getTransactionsAsMerchant() {
        return $this->transactions_as_merchant;
    }

    /**
     * Get the marketplace transactions in which this actor was the customer.
     * @return ArrayCollection
     */
    public function getTransactionsAsCustomer() {
        return $this->transactions_as_customer;
    }

    public function parseAPIRequest(Actor $actor = null, array $data, array $options = []) {
        if (isset($data['email']))
            $this->setEmail($data['email']);
        if (isset($data['ext_user_id']))
            $this->setExternalUserID($data['ext_user_id']);
        if (isset($data['org_name']))
            $this->setOrgName($data['org_name']);
        if (isset($data['dep_name']))
            $this->setDepName($data['dep_name']);
        if (isset($data['full_name']))
            $this->setFullName($data['full_name']);
        if (isset($data['street_address']))
            $this->setStreetAddress($data['street_address']);
        if (isset($data['postal_code']))
            $this->setPostalCode($data['postal_code']);
        if (isset($data['location']))
            $this->setLocation($data['location']);
        if (isset($data['country']))
            $this->setCountry($data['country']);
        if (isset($data['eu_vat_id']))
            $this->setEUVatID($data['eu_vat_id']);
        if (isset($data['vat_exempt']))
            $this->setVATExempt($data['vat_exempt']);
        if (isset($data['company_tax_id']))
            $this->setCompanyTaxID($data['company_tax_id']);
        if (isset($data['cost_center_id']))
            $this->setCostCenterID($data['cost_center_id']);
    }

    public static function isValidAPIRequest(array $data, array $options = []) {
        throw new \Exception("no validation required");
        // TODO: add validation for location
        // TODO: add validation for VAT IDs
        // (list: https://de.wikipedia.org/wiki/Umsatzsteuer-Identifikationsnummer)
    }

    public function createAPIResponse(Actor $actor = null, array $options = []) {
        if (isset($options['_detail'])) {
            Assert::isAOf($options['_detail'], ActorDetail::class);
            Assert::eq($this, $options['_detail']->getActor());
            $detail = $options['_detail'];
        } else
            $detail = $this->getOrCreateCurrentDetail();

        return [
            'email' => $this->getEmail(),
            'ext_user_id' => $this->getExternalUserID(),
            'org_name' => $detail->getOrgName(),
            'dep_name' => $detail->getDepName(),
            'full_name' => $detail->getFullName(),
            'street_address' => $detail->getStreetAddress(),
            'postal_code' => $detail->getPostalCode(),
            'location' => $detail->getLocation(),
            'country' => $detail->getCountry(),
            'eu_vat_id' => $detail->getEUVatID(),
            'vat_exempt' => $detail->isVATExempt(),
            'company_tax_id' => $detail->getCompanyTaxID(),
            'cost_center_id' => $detail->getCostCenterID(),
            'registration_info_line' => null, // TODO: add field for this
            'management_info_line' => null, // TODO: add field for this
            'saved_details_id' => $detail->getId()
        ];
    }

    /**
     * Get the cycle that is used for invoice numbering for invoices
     * sent from this actor as a merchant. One of "Y", "m" or "d".
     * @return string
     */
    public function getInvoicingCycle() : string {
        // TODO: decide whether to store this in DB or load from CO
        return "Y";
    }

    /**
     * Generate a due date for an invoice.
     * 
     * @param Carbon $creationDate The date on which the invoice was created.
     * @return Carbon
     */
    public function getDueDateFromCreationDate(Carbon $creationDate) : Carbon {
        // TODO: decide how actors can have their own due dates
        return $this->getMarketplace()->getDueDateFromCreationDate($creationDate);
    }

    /**
     * Get the CloudObjects AccountGateways Account Context for this actor.
     */
    public function getAccountContext() : AccountContext {
        if (!isset($this->context)) {
            $aauidIri = AAUIDParser::fromString($this->getAAUID());
            $this->context = new AccountContext($aauidIri, $this->access_token);
        }

        return $this->context;
    }

    /**
     * Get the connected account on the marketplace.
     */
    public function getMarketplaceAccount() {
        return $this->getAccountContext()->getConnectedAccountForService(
            'coid://' . $this->getMarketplace()->getServiceDomain()
        );
    }

    public function offsetExists($key) : bool {        
        return true; // always return true and then return empty strings for non-existing fields
    }

    /**
     * @return mixed
     */
    public function offsetGet($key) {
        if (substr($key, 0, 8) == 'account>')
            return $this->getReader()->getFirstValueString(
                $this->getMarketplaceAccount(),
                    substr($key, 8), '');
        if (substr($key, 0, 7) == 'person>')
            return $this->getReader()->getFirstValueString(
                $this->getAccountContext()->getPerson(),
                    substr($key, 7), '');                    
        return '';
    }

    public function offsetSet($key, $value) : void {
        // This should not be implemented and silently fails.
    }

    public function offsetUnset($key) : void {
        // This should not be implemented and silently fails.
    }

    /**
     * Indicates whether this actor is a business, i.e., has an organization name
     * or company tax registration, or not (= an individual consumer).
     * 
     * @return bool
     */
    public function isBusiness() : bool {
        return ($this->getOrgName() != null || $this->getEUVatID() != null || $this->getCompanyTaxID() != null);
    }

    /**
     * Returns all details for the actor.
     */
    public function getDetails() {
        return $this->details;
    }

    /**
     * Return an ActorDetail that represents the current state of the actor;
     * either fetches from existing details or creates new.
     */
    public function getOrCreateCurrentDetail() : ActorDetail {
        foreach ($this->getDetails() as $detail) {
            if ($detail->matchesAddress($this)) {
                // Found existing detail, return it
                return $detail;
            }
        }

        // No detail found, create new
        $newDetail = new ActorDetail($this);
        $this->details->add($newDetail);
        
        return $newDetail;
    }

    /**
     * Get the identity of the actor, or NULL if none.
     * @return Identity|null
     */
    public function getIdentity() : ?Identity {
        return $this->identity;
    }

    /**
     * Set the identity of the actor.
     * @param Identity $identity
     * @return Actor self-reference for chaining
     */
    public function setIdentity(Identity $identity) : Actor {
        $this->identity = $identity;
        return $this;
    }

    /**
     * Get the identity of the actor, or create one.
     * @return Identity
     */
    public function getOrCreateIdentity() : Identity {
        if (!$this->identity) {
            $this->identity = new Identity;
            $this->identity->getActors()->add($this);
        }

        return $this->identity;
    }

}
