<?php

namespace ProgrammaBill\Entities;

use Doctrine\ORM\Mapping AS ORM;
use Doctrine\Common\Collections\ArrayCollection;
use Carbon\Carbon;
use Punic\Territory;
use ProgrammaBill\Interfaces\Address;

/**
 * The ActorDetail refers to a set of details for an actor, allowing management
 * of multiple addresses or other information.
 *
 * @ORM\Entity
 * @ORM\Table(name="mi_actor_details")
 */
class ActorDetail extends BaseEntity implements Address {

    /** @ORM\ManyToOne(targetEntity="Actor") **/
    private $actor;

    /** @ORM\Column(type="string", length=70, nullable=true) **/
    private $org_name;

    /** @ORM\Column(type="string", length=70, nullable=true) **/
    private $dep_name;

    /** @ORM\Column(type="string", length=70, nullable=true) **/
    private $full_name;

    /** @ORM\Column(type="string", length=70, nullable=true) **/
    private $street_address;

    /** @ORM\Column(type="string", length=10, nullable=true) **/
    private $postal_code;

    /** @ORM\Column(type="string", length=70, nullable=true) **/
    private $location;

    /** @ORM\Column(type="string", length=2, nullable=true) **/
    private $country;

    /** @ORM\Column(type="string", length=50, nullable=true) **/
    private $eu_vat_id;

    /** @ORM\Column(type="boolean") */
    private $vat_exempt = false;

    /** @ORM\Column(type="string", length=50, nullable=true) **/
    private $company_tax_id;

    /** @ORM\Column(type="string", length=20, nullable=true) **/
    private $cost_center_id;

    /**
     * @ORM\OneToMany(targetEntity="Transaction", mappedBy="merchant_detail", cascade={"persist"})
     * @ORM\OrderBy({"id" = "DESC"})
     */
    private $transactions_as_merchant;

    /**
     * @ORM\OneToMany(targetEntity="Transaction", mappedBy="customer_detail", cascade={"persist"})
     * @ORM\OrderBy({"id" = "DESC"})
     */
    private $transactions_as_customer;


    /**
     * Creates a new ActorDetail. All information is copied
     * from the current state of the actor.
     * 
     * @param Actor $actor Actor that the details belong to
     */
    public function __construct(Actor $actor) {
        parent::__construct();
        $this->actor = $actor;
        
        // copy details from actor
        $this->org_name = $actor->getOrgName();
        $this->dep_name = $actor->getDepName();
        $this->full_name = $actor->getFullName();
        $this->street_address = $actor->getStreetAddress();
        $this->postal_code = $actor->getPostalCode();
        $this->location = $actor->getLocation();
        $this->country = $actor->getCountry();
        $this->eu_vat_id = $actor->getEUVatID();
        $this->vat_exempt = $actor->isVATExempt();
        $this->company_tax_id = $actor->getCompanyTaxID();
        $this->cost_center_id = $actor->getCostCenterID();

        $this->transactions_as_merchant = new ArrayCollection;
        $this->transactions_as_customer = new ArrayCollection;
    }

    /**
     * Indicates whether this address is an exact match for another.
     * @return bool
     */
    public function matchesAddress(Address $address) : bool {
        return (
            $this->getOrgName() == $address->getOrgName()
            && $this->getDepName() == $address->getDepName()
            && $this->getFullName() == $address->getFullName()
            && $this->getStreetAddress() == $address->getStreetAddress()
            && $this->getPostalCode() == $address->getPostalCode()
            && $this->getLocation() == $address->getLocation()
            && $this->getCountry() == $address->getCountry()
            && $this->getEUVatID() == $address->getEUVatID()
            && $this->isVATExempt() == $address->isVATExempt()
            && $this->getCompanyTaxID() == $address->getCompanyTaxID()
            && $this->getCostCenterID() == $address->getCostCenterID()
        );
    }

    /**
     * Get the actor to which these details belong.
     * @return Actor
     */
    public function getActor() : Actor {
        return $this->actor;
    }

    /**
     * Get the organization name / common display name for these details.
     * @return string|null
     */
    public function getOrgName() : ?string {
        return $this->org_name;
    }

    /**
     * Get the department name / subtitle display name for these details.
     * @return string|null
     */
    public function getDepName() : ?string {
        return $this->dep_name;
    }

    /**
     * Get the full name for these details.
     * @return string|null
     */
    public function getFullName() : ?string {
        return $this->full_name;
    }

    /**
     * Get the street address, typically a street name
     * and house number, for these details.
     * @return string|null
     */
    public function getStreetAddress() : ?string {
        return $this->street_address;
    }

    /**
     * Get the postal code for the location of this actor.
     * @return string|null
     */
    public function getPostalCode() : ?string {
        return $this->postal_code;
    }

    /**
     * Get the name of the location, e.g. city or village,
     * of this actor.
     * @return string|null
     */
    public function getLocation() : ?string {
        return $this->location;
    }

    /**
     * Get the country in which the actor is located
     * as an ISO-3166-1 code.
     * @return string|null
     */
    public function getCountry() : ?string {
        return $this->country;
    }

    /**
     * Indicates whether the actor is located in the European Union.
     * @return bool
     */
    public function isLocatedInEU() : bool {
        return in_array($this->getCountry(), Territory::getChildTerritoryCodes('EU'));
    }

    /**
     * Get the European VAT identification number for these details.
     */
    public function getEUVatID() {
        return $this->eu_vat_id;
    }

    /**
     * Indicates whether this actor is exempt from charging VAT.
     */ 
    public function isVATExempt() : bool {
        return $this->vat_exempt;
    }

    /**
     * Get the company Tax ID for these details.
     * @return string|null
     */ 
    public function getCompanyTaxID() : ?string {
        return $this->company_tax_id;
    }

    /**
     * Get the cost center ID for these details.
     * @return string|null
     */ 
    public function getCostCenterID() : ?string {
        return $this->cost_center_id;
    }

    /**
     * Get the marketplace transactions in which these details was the merchant.
     * @return ArrayCollection
     */
    public function getTransactionsAsMerchant() {
        return $this->transactions_as_merchant;
    }

    /**
     * Get the marketplace transactions in which this actor was the customer.
     * @return ArrayCollection
     */
    public function getTransactionsAsCustomer() {
        return $this->transactions_as_customer;
    }

    /**
     * Generate a due date for an invoice.
     * 
     * @param Carbon $creationDate The date on which the invoice was created.
     * @return Carbon
     */
    public function getDueDateFromCreationDate(Carbon $creationDate) : Carbon {
        // delegate to actor
        return $this->getActor()->getDueDateFromCreationDate($creationDate);
    }

    /**
     * Indicates whether this actor is a business, i.e., has an organization name,
     * or not (= an individual consumer).
     * 
     * @return bool
     */
    public function isBusiness() : bool {
        return ($this->getOrgName() != null || $this->getEUVatID() != null);
    }

}
