<?php

namespace ProgrammaBill\Entities;

use DateTime;
use Carbon\Carbon;
use Webmozart\Assert\Assert;
use Doctrine\ORM\Mapping AS ORM;
use Doctrine\Common\Collections\ArrayCollection;
use ProgrammaBill\Interfaces\APIObject;
use ProgrammaBill\Helpers;
use ProgrammaBill\InvoiceFormatter;
use ProgrammaBill\Repositories\InvoiceRepository;

/**
 * An Invoice is a bill for good, services or related feeds associated with
 * a transaction.
 *
 * @ORM\Entity(repositoryClass="ProgrammaBill\Repositories\InvoiceRepository")
 * @ORM\Table(name="mi_invoices")
 */
class Invoice extends BaseEntity implements APIObject {

    const TYPE_MERCHANT_TO_CUSTOMER = 1;
    const TYPE_PLATFORM_TO_CUSTOMER = 2;
    const TYPE_PLATFORM_TO_MERCHANT = 3;
    const TYPE_MERCHANT_TO_PLATFORM = 4;

    const TYPE_MERCHANT_TO_CUSTOMER_REVERSED = 11;
    const TYPE_PLATFORM_TO_CUSTOMER_REVERSED = 12;
    const TYPE_PLATFORM_TO_MERCHANT_REVERSED = 13;
    const TYPE_MERCHANT_TO_PLATFORM_REVERSED = 14;

    const TAX_DISPLAY_DEFAULT = 0;
    const TAX_DISPLAY_INCLUDE = 1;
    const TAX_DISPLAY_EXCLUDE = 2;

    const ACTOR_PLATFORM = 1;
    const ACTOR_MERCHANT = 2;
    const ACTOR_CUSTOMER = 3;

    /** @ORM\ManyToOne(targetEntity="Transaction", inversedBy="invoices") **/
    private $transaction;

    /** @ORM\Column(type="smallint") **/
    private $invoice_type;

    /** @ORM\Column(type="string", length=15) **/
    private $number;

    /** @ORM\Column(type="json_array") **/
    private $items = [];

    /** @ORM\Column(type="decimal", precision=9, scale=2, options={"default":0}) **/
    private $total_payable_amount = 0;

    /** @ORM\Column(type="string", length=3) **/
    private $currency;

    /** @ORM\Column(type="datetime") **/
    private $due_date;

    /** @ORM\Column(type="boolean") **/
    private $is_payment_received = false;

    /** @ORM\Column(type="boolean") **/
    private $is_paid_out = false;

    /** @ORM\Column(type="smallint", options={"default":0}) */
    private $tax_display = 0;

    /** @ORM\Column(type="boolean") **/
    private $items_include_tax = false;

    /** @ORM\Column(type="json_array") **/
    private $meta = [];

    /** @ORM\ManyToOne(targetEntity="PaymentInstruction") **/
    private $payment_instruction;

    /** @ORM\OneToMany(targetEntity="InvoiceAction", mappedBy="invoice", cascade={"persist"}) **/
    private $actions;
    
    /** @ORM\OneToMany(targetEntity="InvoiceLink", mappedBy="invoice", cascade={"persist"}) **/
    private $links;

    private function invoiceTypeNativeToString($invoice_type) {
        switch ($invoice_type) {
            case self::TYPE_MERCHANT_TO_CUSTOMER:
                return "M2C";
            case self::TYPE_PLATFORM_TO_MERCHANT:
                return "P2M";
            case self::TYPE_PLATFORM_TO_CUSTOMER:
                return "P2C";
            case self::TYPE_MERCHANT_TO_PLATFORM:
                return "M2P";
            case self::TYPE_MERCHANT_TO_CUSTOMER_REVERSED:
                return "M2C-R";
            case self::TYPE_PLATFORM_TO_MERCHANT_REVERSED:
                return "P2M-R";
            case self::TYPE_PLATFORM_TO_CUSTOMER_REVERSED:
                return "P2C-R";
            case self::TYPE_MERCHANT_TO_PLATFORM_REVERSED:
                return "M2P-R";
            default:
                return null;
        }
    }

    private function invoiceTypeStringToNative($invoice_type) {
        switch ($invoice_type) {
            case "M2C":
                return self::TYPE_MERCHANT_TO_CUSTOMER;
            case "P2M":
                return self::TYPE_PLATFORM_TO_MERCHANT;
            case "P2C":
                return self::TYPE_PLATFORM_TO_CUSTOMER;
            case "M2P":
                return self::TYPE_MERCHANT_TO_PLATFORM;
            case "M2C-R":
                return self::TYPE_MERCHANT_TO_CUSTOMER_REVERSED;
            case "P2M-R":
                return self::TYPE_PLATFORM_TO_MERCHANT_REVERSED;
            case "P2C-R":
                return self::TYPE_PLATFORM_TO_CUSTOMER_REVERSED;
            case "M2P-R":
                return self::TYPE_MERCHANT_TO_PLATFORM_REVERSED;
            default:
                return null;
        }
    }

    private function inverseInvoiceType($invoice_type) {
        switch ($invoice_type) {
            case self::TYPE_MERCHANT_TO_CUSTOMER:
                return self::TYPE_MERCHANT_TO_CUSTOMER_REVERSED;
            case self::TYPE_PLATFORM_TO_MERCHANT:
                return self::TYPE_PLATFORM_TO_MERCHANT_REVERSED;
            case self::TYPE_PLATFORM_TO_CUSTOMER:
                return self::TYPE_PLATFORM_TO_CUSTOMER_REVERSED;
            case self::TYPE_MERCHANT_TO_PLATFORM:
                return self::TYPE_MERCHANT_TO_PLATFORM_REVERSED;
            case self::TYPE_MERCHANT_TO_CUSTOMER_REVERSED:
                return self::TYPE_MERCHANT_TO_CUSTOMER;
            case self::TYPE_PLATFORM_TO_MERCHANT_REVERSED:
                return self::TYPE_PLATFORM_TO_MERCHANT;
            case self::TYPE_PLATFORM_TO_CUSTOMER_REVERSED:
                return self::TYPE_PLATFORM_TO_CUSTOMER;
            case self::TYPE_MERCHANT_TO_CUSTOMER_REVERSED:
                return self::TYPE_MERCHANT_TO_PLATFORM;
            default:
                return null;
        }
    }

    private function taxDisplayNativeToString($tax_display) {
        switch ($tax_display) {
            case self::TAX_DISPLAY_DEFAULT:
                return "default";
            case self::TAX_DISPLAY_INCLUDE:
                return "include";
            case self::TAX_DISPLAY_EXCLUDE:
                return "exclude";
            default:
                return null;
        }
    }

    private function taxDisplayStringToNative($tax_display) {
        switch ($tax_display) {
            case "default":
                return self::TAX_DISPLAY_DEFAULT;
            case "include":
                return self::TAX_DISPLAY_INCLUDE;
            case "exclude":
                return self::TAX_DISPLAY_EXCLUDE;
            default:
                return null;
        }
    }

    private function normalizeRate($rate) {
        Assert::true($rate == 'R' || is_numeric($rate), "Tax rate must be 'R' or a numeric value.");
        return is_numeric($rate) ? (float)$rate : $rate;
    }

    private function updateTotalPayableAmount() {
        $amount = 0;
        foreach ($this->getTaxRates() as $rate)
            $amount += $this->getPayableAmountForRate($rate);
        
        $this->total_payable_amount = round($amount, 2);
    }

    /**
     * Creates a new invoice.
     * @param Transaction $transaction The transaction to which this invoice belongs.
     * @param int $invoice_type A valid invoice type
     * @param string $number The invoice number.
     * @param bool $items_include_tax Indicate whether items specified in items already include taxes.
     */
    public function __construct(Transaction $transaction, int $invoice_type, string $number, bool $items_include_tax = false) {
        Assert::notNull($this->invoiceTypeNativeToString($invoice_type));

        parent::__construct();
        $this->transaction = $transaction;
        $this->invoice_type = $invoice_type;
        $this->number = $number;
        $this->items_include_tax = $items_include_tax;
        $this->actions = new ArrayCollection;
    }

    /**
     * Get the transaction to which this invoice belongs.
     * @return Transaction
     */
    public function getTransaction() {
        return $this->transaction;
    }

    /**
     * Get the marketplace to which this invoice belongs.
     * @return Marketplace
     */
    public function getMarketplace() {
        return $this->getTransaction()->getMarketplace();
    }

    /**
     * Get the type of this invoice.
     * @return integer
     */
    public function getInvoiceType() {
        return $this->invoice_type;
    }

    /**
     * Get the type of this invoice.
     * @return integer
     */
    public function getInvoiceTypeAsString() {
        return $this->invoiceTypeNativeToString($this->invoice_type);
    }

    /**
     * Get the merchant- or platform-specific invoice number for this invoice.
     * @return string
     */
    public function getNumber() {
        return $this->number;
    }

    /**
     * Indicates whether items in this invoice include tax or not.
     */
    public function doItemsIncludeTax() {
        return $this->items_include_tax;
    }

    /**
     * Get the items listed on this invoice.
     * @return array
     */
    public function getItems() {
        return $this->items;
    }

    /**
     * Removes all items from this invoice and sets the tax inclusion setting.
     * @param bool $items_include_tax Indicate whether items specified in items already include taxes.
     * @return Invoice self-reference for chaining
     */
    public function resetItems(bool $items_include_tax = false) {
        $this->items = [];
        $this->total_payable_amount = 0;
        $this->items_include_tax = $items_include_tax;
        return $this;
    }

    /**
     * Add a new item to this invoice.
     *
     * @param string $name Name of the item.
     * @param float $unit_price Price per single item in invoice currency, without taxes.
     * @param float $count Number of units for this item. Defaults to 1.
     * @param string $unit_name Unit of measurement for this item. Only used with count > 1.
     * @param float|string $tax_rate Rate for VAT or sales tax in percentage points. Defaults to 0.
     */
    public function addItem(string $name, float $unit_price, float $count = 1, string $unit_name = '', $tax_rate = 0) {
        $tax_rate = $this->normalizeRate($tax_rate);

        if ($tax_rate != 0)
            Assert::false($this->getSender()->isVATExempt(),
                "VAT-exempt senders can only specify a tax_rate of 0 for invoice items.");

        $this->items[] = [
            'name' => $name,
            'unit_price' => $unit_price,
            'unit_name' => $unit_name,
            'count' => $count,
            'tax_rate' => is_numeric($tax_rate) ? (float)$tax_rate : $tax_rate
        ];
        $this->updateTotalPayableAmount();
        return $this;
    }

    /**
     * Get the total payable amount of this invoice in the specified currency.
     * This amount contains any required sales tax or VAT if applicable.
     * @return float
     */
    public function getTotalPayableAmount() {
        return (float)$this->total_payable_amount;
    }


    /**
     * Get the currency of the invoice as an ISO 4217 code.
     * @return string
     */
    public function getCurrency() {
        return $this->currency;
    }

    /**
     * Get the currency of the invoice as an ISO 4217 code.
     *
     * @param string $currency
     * @return Invoice self-reference for chaining
     */
    public function setCurrency(string $currency) {
        Assert::length($currency, 3); // TODO: validate against a list of codes
        $this->currency = $currency;
        return $this;
    }

    /**
     * Get the date on which this invoice is due to be paid.
     * @return DateTime
     */
    public function getDueDate() {
        return $this->due_date;
    }

    /**
     * Set the date on which this invoice is due to be paid.
     * @param DateTime $due_date
     * @return Invoice self-reference for chaining
     */
    public function setDueDate(DateTime $due_date) {
        $this->due_date = $due_date;
        return $this;
    }

    /**
     * Indicates whether payment for this invoice was received or not.
     * @return boolean
     */
    public function isPaymentReceived() {
        return $this->is_payment_received;
    }

    /**
     * Specifies whether payment for this invoice was received or not.
     * @param boolean $is_payment_received
     * @return Invoice self-reference for chaining
     */
    public function setPaymentReceived(bool $is_payment_received) {
        $this->is_payment_received = $is_payment_received;
        return $this;
    }

    /**
     * Indicates whether payment for this invoice was received by the merchant.
     * Only used with TYPE_MERCHANT_TO_CUSTOMER.
     * @return boolean
     */
    public function isPaidOut() {
        return $this->is_paid_out;
    }

    /**
     * Specifies whether payment for this invoice was received by the merchant.
     * Only used with TYPE_MERCHANT_TO_CUSTOMER.
     * Throws exception when set to true for other types.
     * @param boolean $is_paid_out
     * @return Invoice self-reference for chaining
     */
    public function setPaidOut(bool $is_paid_out) {
        if ($is_paid_out == true)
            Assert::eq($this->invoice_type, self::TYPE_MERCHANT_TO_CUSTOMER, "Type 'paid_out' only applicable to merchant invoices.");
        $this->is_paid_out = $is_paid_out;
        return $this;
    }

    /**
     * Returns how taxes should be displayed on this invoice.
     */
    public function getTaxDisplay() {
        return $this->tax_display;
    }

    /**
     * Get user-supplied meta-data for this invoice.
     * @return array
     */
    public function getMeta() {
        return $this->meta;
    }

    /**
     * Set user-supplied meta-data for this invoice.
     *
     * @param array $user_meta
     * @return Invoice self-reference for chaining
     */
    public function setMeta(array $meta) {
        $this->meta = $meta;
        return $this;
    }

    /**
     * Get the payment instruction for this invoice or NULL if none specified.
     * @return PaymentInstruction
     */ 
    public function getPaymentInstruction() {
        return $this->payment_instruction;
    }

    /**
     * Get all the actions taken against this invoice after creation.
     */
    public function getActions() {
        return $this->actions;
    }

    /**
     * Find all actions with the provided type.
     */
    public function findActionsWithType(int $actionType) {
        return $this->getActions()->filter(function($action) use ($actionType) {
            return ($action->getActionType() == $actionType);
        });
    }

    /**
     * Get all the remote links for this invoice.
     */
    public function getLinks() {
        return $this->links;
    }

    /**
     * Get all tax rates that were used for at least one item.
     * @return array
     */
    public function getTaxRates() {
        $rates = [];
        foreach ($this->items as $i)
            if (!in_array($i['tax_rate'], $rates))
                $rates[] = $i['tax_rate'];
        return $rates;
    }

    /**
     * Calculates the net sum of items before taxes.
     * @return float
     */
    public function getTotalPretaxAmount() {
        $amount = 0;
        if ($this->items_include_tax) {
            // Taxes are included, so we need to remove them
            foreach ($this->items as $i)
                $amount += ($i['tax_rate'] === 'R'
                    ? round($i['unit_price'] * $i['count'], 2) // if reverse charged, same as total
                    : round($i['unit_price'] * $i['count'] / (1 + $i['tax_rate']/100), 2)
                );
        } else {
            // Taxes are excluded, so we can take the original prices
            foreach ($this->items as $i)
                $amount += round($i['unit_price'] * $i['count'], 2);
        }
        return $amount;
    }

    /**
     * Calculates the sum of all taxes.
     * @return float
     */
    public function getTotalTaxAmount() {
        $amount = 0;
        foreach ($this->getTaxRates() as $tr)
            $amount += $this->getTaxAmountAtRate($tr);
        return round($amount, 2);
    }

    /**
     * Calculates the sum of taxes for all items taxed at a specific rate.
     * @param float $rate The tax rate used for calculation.
     * @return float
     */
    public function getTaxAmountAtRate($rate) {
        $rate = $this->normalizeRate($rate);

        $amount = 0;
        if (!in_array($rate, [ 0, 'R'])) {
            if ($this->items_include_tax) {
                // Taxes are included, so we need to take them out of the price
                foreach ($this->items as $i)
                    if ($i['tax_rate'] == $rate)
                        $amount += round(($i['unit_price'] * $i['count']) - ($i['unit_price'] * $i['count'] / (1 + $rate/100)), 2);
            } else {
                // Taxes are excluded, so we can calculate them off the price
                foreach ($this->items as $i)
                    if ($i['tax_rate'] == $rate)
                        $amount += round($i['unit_price'] * $i['count'] * $rate/100, 2);
            }
        }
        
        return $amount;
    }

    /**
     * Calculates the total pre-tax amount for all items taxed at a specific rate.
     * @param float $rate The tax rate used for calculation.
     * @return float
     */
    public function getPretaxAmountForRate($rate) {
        $rate = $this->normalizeRate($rate);

        $amount = 0;
        if ($this->items_include_tax) {
            // Taxes are included, so we need to take them out of the price
            foreach ($this->items as $i)
                if ($i['tax_rate'] === $rate)
                    $amount += ($i['tax_rate'] === 'R'
                        ? round($i['unit_price'] * $i['count'], 2) // if reverse charged, same as no tax
                        : round($i['unit_price'] * $i['count'] / (1 + $i['tax_rate'] / 100), 2)
                    );
        } else {
            // Taxes are excluded, so we can just use the base price
            foreach ($this->items as $i)
                if ($i['tax_rate'] === $rate)
                    $amount += round($i['unit_price'] * $i['count'], 2);
        }
        return $amount;
    }

    /**
     * Get the unit price for an item without tax, if applicable.
     * @param array $item Item
     * @return float
     */
    public function getPretaxUnitAmountForItem(array $item) {
        if ($this->items_include_tax) {
            // Taxes are included, so we need to take them out of the price
            return round($item['tax_rate'] === 'R'
                ? $item['unit_price']
                : $item['unit_price'] / (1 + $item['tax_rate'] / 100), 2);
        } else {
            // Taxes are excluded, so we just take the item price without changes
            return $item['unit_price'];
        }
    }

    /**
     * Get the subtotal for an item without tax, if applicable.
     * @param array $item Item
     * @return float
     */
    public function getPretaxSubtotalAmountForItem(array $item) {
        if ($this->items_include_tax) {
            // Taxes are included, so we need to take them out of the price
            return round($item['tax_rate'] === 'R'
                ? $item['count'] * $item['unit_price']
                : $item['count'] * $item['unit_price'] / (1 + $item['tax_rate'] / 100), 2);
        } else {
            // Taxes are excluded, so we just take the item price without changes
            return round($item['count'] * $item['unit_price'], 2);
        }        
    }

    /**
     * Get the unit price for an item including tax, if applicable.
     * @param array $item Item
     * @return float
     */
    public function getPostTaxUnitAmountForItem(array $item) {
        if ($this->items_include_tax) {
            // Taxes are included, so we can take the item price without changes
            return $item['unit_price'];
        } else {
            // Taxes are excluded, so we add them on top of the base price
            return round($item['tax_rate'] === 'R'
                ? $item['unit_price']
                : $item['unit_price'] * (1 + $item['tax_rate'] / 100), 2);
        }
    }

    /**
     * Get the subtotal for an item including tax, if applicable.
     * @param array $item Item
     * @return float
     */
    public function getPostTaxSubtotalAmountForItem(array $item) {
        if ($this->items_include_tax) {
            // Taxes are included, so we can take the item price without changes
            return round($item['count'] * $item['unit_price'], 2);
        } else {
            // Taxes are excluded, so we add them on top of the base price
            return round($item['tax_rate'] === 'R'
                ? $item['count'] * $item['unit_price']
                : $item['count'] * $item['unit_price'] * (1 + $item['tax_rate'] / 100), 2);
        }
    }

    /**
     * Get the unit tax amount for an item.
     * @param array $item Item
     * @return float
     */
    public function getTaxUnitAmountForItem(array $item) {
        return round($this->getPostTaxUnitAmountForItem($item)
            - $this->getPretaxUnitAmountForItem($item), 2);
    }

    /**
     * Get the subtotal tax amount for an item.
     * @param array $item Item
     * @return float
     */
    public function getTaxSubtotalAmountForItem(array $item) {
        return round($this->getPostTaxSubtotalAmountForItem($item)
            - $this->getPretaxSubtotalAmountForItem($item), 2);
    }

    /**
     * Calculates the payable amount including tax for all items taxed at a specific rate.
     * @param float $rate The tax rate used for calculation.
     * @return float
     */
    public function getPayableAmountForRate($rate) {
        return round($this->getTaxAmountAtRate($rate)
            + $this->getPretaxAmountForRate($rate), 2);
    }

    /**
     * If this is an original invoice that was reversed, returns the reverse invoice.
     * @return Invoice|null
     */
    public function getReverseInvoice() {
        switch ($this->getInvoiceType()) {
            case self::TYPE_MERCHANT_TO_CUSTOMER:
                return $this->getTransaction()->getInvoiceWithType(self::TYPE_MERCHANT_TO_CUSTOMER_REVERSED);
            case self::TYPE_PLATFORM_TO_MERCHANT:
                return $this->getTransaction()->getInvoiceWithType(self::TYPE_PLATFORM_TO_MERCHANT_REVERSED);
            case self::TYPE_PLATFORM_TO_CUSTOMER:
                return $this->getTransaction()->getInvoiceWithType(self::TYPE_PLATFORM_TO_CUSTOMER_REVERSED);
            case self::TYPE_MERCHANT_TO_PLATFORM:
                return $this->getTransaction()->getInvoiceWithType(self::TYPE_MERCHANT_TO_PLATFORM_REVERSED);
            default:
                return null;
        }
    }

    /**
     * If this is a reverse invoice, returns the original invoice that was reversed.
     * @return Invoice|null
     */
    public function getReversedOriginalInvoice() {
        switch ($this->getInvoiceType()) {
            case self::TYPE_MERCHANT_TO_CUSTOMER_REVERSED:
                return $this->getTransaction()->getInvoiceWithType(self::TYPE_MERCHANT_TO_CUSTOMER);
            case self::TYPE_PLATFORM_TO_MERCHANT_REVERSED:
                return $this->getTransaction()->getInvoiceWithType(self::TYPE_PLATFORM_TO_MERCHANT);
            case self::TYPE_PLATFORM_TO_CUSTOMER_REVERSED:
                return $this->getTransaction()->getInvoiceWithType(self::TYPE_PLATFORM_TO_CUSTOMER);
            case self::TYPE_MERCHANT_TO_PLATFORM_REVERSED:
                return $this->getTransaction()->getInvoiceWithType(self::TYPE_MERCHANT_TO_PLATFORM);
            default:
                return null;
        }
    }

    /**
     * If this is an original invoice that was reversed, returns
     * the number of the reverse invoice, null otherwise.
     * @return string|null
     */
    public function getReverseNumber() {
        $invoice = $this->getReverseInvoice();
        return isset($invoice) ? $invoice->getNumber() : null;
    }

    /**
     * If this is a reverse invoice, returns the number
     * of the original invoice that was reversed, null otherwise.
     * @return string|null
     */
    public function getReversedOriginalNumber() {
        $invoice = $this->getReversedOriginalInvoice();
        return isset($invoice) ? $invoice->getNumber() : null;
    }

    public function parseAPIRequest(Actor $actor = null, array $data, array $options = []) {
        $this->setCurrency($data['currency']);

        if (isset($data['stage']) && $data['stage'] == 'paid_in') {
            // Invoice was already paid, create an action for that
            $this->getActions()->add(new InvoiceAction($this, InvoiceAction::TYPE_PAYIN));
            $this->setPaymentReceived(true);
        } else
        if (isset($data['stage']) && $data['stage'] == 'paid_out') {
            // Invoice was already paid, create an action for that
            $this->getActions()->add(new InvoiceAction($this, InvoiceAction::TYPE_PAYDIRECT));
            $this->setPaymentReceived(true);
            $this->setPaidOut(true);
        }

        if (isset($data['due_date'])) {
            // Due date was specified with request and must be validated
            Assert::numeric(strtotime($data['due_date']), 'The due_date must be a valid date.');
            $dueDate = new DateTime($data['due_date']);
            Assert::greaterThan($data['due_date'], Carbon::now(), 'The due_date must be in the future.');
            $this->setDueDate($dueDate);
        } elseif ($this->isPaymentReceived()) {
            // Payment was received and due date is set to the delivery date
            $this->setDueDate($this->getTransaction()->getDeliveryDate());
        } else {
            // Due date was not specified and is set to default
            $this->setDueDate($this->getSender()->getDueDateFromCreationDate(Carbon::now()));        
        }

        if (isset($data['items_include_tax']))
            $this->resetItems($data['items_include_tax']);
        else
            $this->resetItems();

        foreach ($data['items'] as $i)
            $this->addItem($i['name'], $i['unit_price'],
                @$i['count'], @$i['unit_name'], @$i['tax_rate']);

        if (isset($data['meta']))
            $this->setMeta($data['meta']);
        
        if (isset($data['payment_instruction'])
                && isset($options['_payment_instruction_repository']))
            $this->payment_instruction = $options['_payment_instruction_repository']
                ->getWithCOIDString($data['payment_instruction']);

        if (isset($data['tax_display']))
            $this->tax_display = $this->taxDisplayStringToNative($data['tax_display']);
    }

    public static function isValidAPIRequest(array $data, array $options = []) {
        if (!Helpers::containsAllFields($data, [ 'type', 'currency', 'items' ])
            || !is_array($data['items']))
        return false;

        if (isset($data['stage']))
            Assert::oneOf($data['stage'], [ 'due', 'overdue', 'paid_in', 'paid_out' ], 'Invalid type for stage.');

        if (isset($data['tax_display']))
            Assert::oneOf($data['tax_display'], [ 'default', 'include', 'exclude' ], 'Invalid type for tax_display.');

        foreach ($data['items'] as $i)
            if (!Helpers::containsAllFields($i, [ 'name', 'unit_price' ]))
                return false;

        return true;
    }

    /**
     * Indicates whether this is an outbound invoice from the perspective of
     * a certain actor i.e. if that actor is the merchant and this invoice
     * was created on behalf of them. If actor is passed as NULL, takes the
     * platform perspective.
     *
     * @param Actor $actor
     * @return boolean
     */
    public function isOutbound(Actor $actor = null) {
        if (!isset($actor)) {
            // Platform perspective
            return ($this->getSenderType() == self::ACTOR_PLATFORM);
        } else {
            // Merchant/customer perspective
            return ($actor == $this->getTransaction()->getMerchant()
                && $this->getInvoiceType() == self::TYPE_MERCHANT_TO_CUSTOMER);
        }
    }

    /**
     * Get the type of sender, either ACTOR_PLATFORM or ACTOR_MERCHANT.
     */
    public function getSenderType() {
        switch ($this->getInvoiceType()) {
            case self::TYPE_PLATFORM_TO_MERCHANT:
            case self::TYPE_PLATFORM_TO_MERCHANT_REVERSED:
            case self::TYPE_PLATFORM_TO_CUSTOMER:
            case self::TYPE_PLATFORM_TO_CUSTOMER_REVERSED:
                return self::ACTOR_PLATFORM;
            case self::TYPE_MERCHANT_TO_CUSTOMER:
            case self::TYPE_MERCHANT_TO_CUSTOMER_REVERSED:
            case self::TYPE_MERCHANT_TO_PLATFORM:
            case self::TYPE_MERCHANT_TO_PLATFORM_REVERSED:
                return self::ACTOR_MERCHANT;
        }
    }

    /**
     * Get the sender of this invoice. Can either be the merchant or the marketplace.
     * 
     * @return Address
     */
    public function getSender() {
        switch ($this->getSenderType()) {
            case self::ACTOR_PLATFORM:
                return $this->getMarketplace();
            case self::ACTOR_MERCHANT:
                return $this->getTransaction()->getMerchant();
        }
    }

    /**
     * Get the type of receiver, either ACTOR_PLATFORM, ACTOR_MERCHANT or ACTOR_CUSTOMER.
     */
    public function getReceiverType() {
        switch ($this->getInvoiceType()) {
            case self::TYPE_PLATFORM_TO_MERCHANT:
            case self::TYPE_PLATFORM_TO_MERCHANT_REVERSED:
                return self::ACTOR_MERCHANT;
            case self::TYPE_PLATFORM_TO_CUSTOMER:
            case self::TYPE_PLATFORM_TO_CUSTOMER_REVERSED:
            case self::TYPE_MERCHANT_TO_CUSTOMER:
            case self::TYPE_MERCHANT_TO_CUSTOMER_REVERSED:
                return self::ACTOR_CUSTOMER;
            case self::TYPE_MERCHANT_TO_PLATFORM:
            case self::TYPE_MERCHANT_TO_PLATFORM_REVERSED:
                return self::ACTOR_PLATFORM;
        }
    }

    /**
     * Get the receiver of this invoice. Can either be the customer or the merchant.
     * 
     * @return Address
     */
    public function getReceiver() {
        switch ($this->getReceiverType()) {
            case self::ACTOR_MERCHANT:
                return $this->getTransaction()->getMerchant();
            case self::ACTOR_CUSTOMER:
                return $this->getTransaction()->getCustomer();
            case self::ACTOR_PLATFORM:
                return $this->getMarketplace();
        }
    }

    /**
     * Validates if an Actor is involved in the invoice as either sender or receiver.
     */
    public function involvesActor(Actor $actor) {
        return ($this->getSender() == $actor || $this->getReceiver() == $actor);
    }

    /**
     * Indicates whether this is a domestic invoice, i.e., sender and receiver
     * are located in the same country.
     */
    public function isDomesticInvoice() {
        return ($this->getSender()->getCountry() == $this->getReceiver()->getCountry());
    }

    /**
     * Indicates whether this is an invoice between different countries in the European Union.
     */
    public function isEUSingleMarketInvoice() {
        return (!$this->isDomesticInvoice()
            && $this->getSender()->isLocatedInEU()
            && $this->getReceiver()->isLocatedInEU());
    }

    /**
     * Indicates whether this is a foreign invoice between two countries that are
     * not part of a single market like the European Union.
     */
    public function isForeignInvoice() {
        return (!$this->isDomesticInvoice() && !$this->isEUSingleMarketInvoice());
    }

    public function createAPIResponse(Actor $actor = null, array $options = []) {        
        $options = array_merge([
            'embed_items' => false,
            'embed_tax_summary' => false,
            'embed_addresses' => false,
            'embed_transaction_dates_only' => false,
            'embed_actions' => false,
            'embed_links' => false,
            'embed_transaction' => false
        ], $options);

        $now = new DateTime;

        $response = [
            'number' => $this->getNumber(),
            'type' => $this->getInvoiceTypeAsString(),
            'reverses_number' => $this->getReversedOriginalNumber(),
            'currency' => $this->getCurrency(),
            'total_pretax_amount' => $this->getTotalPretaxAmount(),
            'total_payable_amount' => $this->getTotalPayableAmount(),
            'vat_reverse_charged' => ($this->getPretaxAmountForRate('R') > 0),
            'due_date' => $this->getDueDate()->format('Y-m-d'),
            'direction' => $this->isOutbound($actor)
                ? 'outbound' : 'inbound',
            'stage' => !isset($actor) || $this->isOutbound($actor)
                ? ($this->isPaidOut() ? 'paid_out' : ($this->isPaymentReceived() ? 'paid_in' : ($now > $this->getDueDate() ? 'overdue' : 'due')))
                : ($this->isPaymentReceived() ? 'paid' : ($now > $this->getDueDate() ? 'overdue' : 'due')),
            'payment_instruction' => ($this->getPaymentInstruction() != null)
                ? str_replace('coid://marketplace-invoicing.cloudobjects.io/', 'mi:', $this->getPaymentInstruction()->getCOID())
                : null,
            'tax_display' => $this->taxDisplayNativeToString($this->getTaxDisplay()),
            'meta' => $this->getMeta()
        ];

        if (isset($options['_invoice_formatter'])) {
            $response['pdf_url'] = $options['_invoice_formatter']->getSignedPDFURL($this);
        }
        
        if ($options['embed_items']) {
            $response['items'] = [];
            $response['items_include_tax'] = $this->items_include_tax;
            foreach ($this->getItems() as $i) {
                $response['items'][] = array_merge($i, [
                    'unit_price_without_tax' => $this->getPretaxUnitAmountForItem($i),
                    'unit_price_with_tax' => $this->getPostTaxUnitAmountForItem($i),
                    'total_price' => round($i['unit_price'] * $i['count'] ,2),
                    'total_price_without_tax' => round($this->getPretaxUnitAmountForItem($i) * $i['count'], 2),
                    'total_price_with_tax' => round($this->getPostTaxUnitAmountForItem($i) * $i['count'], 2)
                ]);
            }
        }

        if ($options['embed_tax_summary']) {
            $response['tax_summary'] = [];
            foreach ($this->getTaxRates() as $r) {
                $response['tax_summary'][] = [
                    'rate' => $r,
                    'tax_amount' => $this->getTaxAmountAtRate($r),
                    'pretax_amount' => $this->getPretaxAmountForRate($r),
                    'total_amount' => $this->getPayableAmountForRate($r)
                ];
            }
        }

        if ($options['embed_addresses']) {
            $response['sender'] = $this->getSender()
                ->createAPIResponse($actor, $options);
            $response['receiver'] = $this->getReceiver()
                ->createAPIResponse($actor, $options);
        }

        if ($options['embed_transaction']) {
            $response['transaction'] = $this->getTransaction()->createAPIResponse($actor,
                array_merge($options, [ 'list_invoices' => false ]));
        } else
        if ($options['embed_transaction_dates_only']) {
            $t = $this->getTransaction();
            $response['transaction'] = [
                'contract_id' => $t->getContractID(),
                'delivery_date' => $t->getDeliveryDate(),
                'delivery_date_end' => $t->getDeliveryDateEnd(),
                'delivery_date_is_period' => ($t->getDeliveryDateEnd() != null),
                'created_timestamp' => $t->getCreatedTimestamp()
            ];
        }

        if ($options['embed_actions']) {
            $response['actions'] = [];
            foreach ($this->getActions() as $action) {
                $response['actions'][] = $action->createAPIResponse($actor, $options);
            }
        }

        if ($options['embed_links']) {
            $response['links'] = [];
            foreach ($this->getLinks() as $link) {
                $response['links'][] = $link->createAPIResponse($actor, $options);
            }
        }

        return $response;
    }

    /**
     * Reverses the invoice. Returns the reverse invoice object
     * or null if reversal wasn't successful or already reversed.
     * 
     * @param InvoiceRepository $invoiceRepository
     * 
     * @return Invoice|null
     */
    public function reverse(InvoiceRepository $invoiceRepository) {
        // Assert invoice is reversable
        if (!in_array($this->getInvoiceType(), [ self::TYPE_MERCHANT_TO_CUSTOMER, self::TYPE_PLATFORM_TO_MERCHANT,
                self::TYPE_PLATFORM_TO_CUSTOMER, self::TYPE_MERCHANT_TO_PLATFORM]))
            return false;

        // Assert invoice wasn't yet reversed
        if (count($this->findActionsWithType(InvoiceAction::TYPE_REVERSE)) > 0)
            return false;

        // Add reversal action
        $action = new InvoiceAction($this, InvoiceAction::TYPE_REVERSE);
        $this->getActions()->add($action);

        $transaction = $this->getTransaction();

        // Create reverse invoice
        $reverseInvoiceType = $this->inverseInvoiceType($this->getInvoiceType());
        $reverseInvoice = ($this->getSender() == $transaction->getMerchant())
            ? $invoiceRepository->createNewInvoiceForMerchant($transaction, $reverseInvoiceType)
            : $invoiceRepository->createNewInvoiceForMarketplace($transaction, $reverseInvoiceType);
        
        // Copy/initialize fields
        $reverseInvoice->items = $this->items;
        $reverseInvoice->total_payable_amount = -1 * $this->total_payable_amount;
        $reverseInvoice->currency = $this->currency;
        $reverseInvoice->tax_display = $this->tax_display;

        // The following assumes that no payment has taken place so the
        // reverse invoice was settled - TODO: implement payment flow
        $reverseInvoice->due_date = new DateTime;
        $reverseInvoice->is_payment_received = true;
        $reverseInvoice->is_paid_out = true;

        $transaction->getInvoices()->add($reverseInvoice);

        return $reverseInvoice;
    }

}