<?php

namespace ProgrammaBill\Entities;

use Doctrine\ORM\Mapping AS ORM;
use Webmozart\Assert\Assert;
use ProgrammaBill\Interfaces\APIObject;

/**
 * An InvoiceAction is an action taken against an invoice after it was created.
 *
 * @ORM\Entity @ORM\Table(name="mi_invoice_actions")
 */
class InvoiceAction extends BaseEntity implements APIObject {

    const TYPE_PAYIN = 1;
    const TYPE_PAYOUT = 2;
    const TYPE_PAYDIRECT = 3;
    const TYPE_REVERSE = 4;
    const TYPE_NOTIFIED_OVERDUE = 5;
    const TYPE_NOTIFIED_IN_DEFAULT = 6;

    /** @ORM\ManyToOne(targetEntity="Invoice") **/
    private $invoice;

    /** @ORM\Column(type="smallint") **/
    private $action_type;

    /** @ORM\Column(type="json_array") **/
    private $meta = [];

    public static function actionTypeNativeToString($action_type) {
        switch ($action_type) {
            case self::TYPE_PAYIN:
                return "payin";
            case self::TYPE_PAYOUT:
                return "payout";
            case self::TYPE_PAYDIRECT:
                return "paydirect";
            case self::TYPE_REVERSE:
                return "reverse";
            case self::TYPE_NOTIFIED_OVERDUE:
                return "notified_overdue";
            case self::TYPE_NOTIFIED_IN_DEFAULT:
                return "notified_in_default";
            default:
                return null;
        }
    }

    public static function actionTypeStringToNative($action_type) {
        switch ($action_type) {
            case "payin":
                return self::TYPE_PAYIN;
            case "payout":
                return self::TYPE_PAYOUT;
            case "paydirect":
                return self::TYPE_PAYDIRECT;
            case "reverse":
                return self::TYPE_REVERSE;
            case "notified_overdue":
                return self::TYPE_NOTIFIED_OVERDUE;
            case "notified_in_default":
                return self::TYPE_NOTIFIED_IN_DEFAULT;
            default:
                return null;
        }
    }

    public static function allTypes() {
        return [ self::TYPE_PAYIN, self::TYPE_PAYOUT, self::TYPE_PAYDIRECT, self::TYPE_REVERSE, self::TYPE_NOTIFIED_OVERDUE ];
    }

    public function __construct(Invoice $invoice, $action_type = null, array $meta = null) {
        parent::__construct();
        $this->invoice = $invoice;
        if ($action_type != null) {
            Assert::oneOf($action_type, self::allTypes());
            $this->action_type = $action_type;
        }
        if ($meta != null) {
            $this->meta = $meta;
        }
    }

    /**
     * Get the invoice this action is for.
     */
    public function getInvoice() {
        return $this->invoice;
    }

    /**
     * Get the type of the action.
     */
    public function getActionType() {
        return $this->action_type;
    }

    /**
     * Get user-supplied meta-data for this action.
     * @return array
     */
    public function getMeta() {
        return $this->meta;
    }

    public function parseAPIRequest(Actor $actor = null, array $data, array $options = []) {
        $this->action_type = self::actionTypeStringToNative($data['action_type']);
        if (isset($data['meta']))
            $this->meta = $data['meta'];
    }

    public static function isValidAPIRequest(array $data, array $options = []) {
        return (isset($data['action_type']) &&
            self::actionTypeStringToNative($data['action_type']) != null);
    }

    public function createAPIResponse(Actor $actor = null, array $options = []) {
        return [
            'action_type' => self::actionTypeNativeToString($this->getActionType()),
            'created_timestamp' => $this->getCreatedTimestamp()->format('c'),
            'meta' => $this->meta
        ];
    }

    public function validate() {
        switch ($this->getActionType()) {
            case InvoiceAction::TYPE_PAYIN:
                Assert::false($this->invoice->isPaymentReceived(), "Already received payment.");
                break;
            case InvoiceAction::TYPE_PAYOUT:
                Assert::true($this->invoice->isPaymentReceived(), "Invoice must have a payin before it can be market as paid out.");
                Assert::false($this->invoice->isPaidOut(), "Already paid out.");
                break;
            case InvoiceAction::TYPE_PAYDIRECT:
                Assert::false($this->invoice->isPaymentReceived(), "Already received payment.");
                Assert::false($this->invoice->isPaidOut(), "Already paid out.");
                break;
            case InvoiceAction::TYPE_NOTIFIED_OVERDUE:
            case InvoiceAction::TYPE_NOTIFIED_IN_DEFAULT:
                // no validation
                break;
            default:
                throw new Exception("Invalid action type.");
        }
    }

    public function updateInvoiceStatus() {
        $this->validate();

        switch ($this->getActionType()) {
            case InvoiceAction::TYPE_PAYIN:
                $this->invoice->setPaymentReceived(true);
                break;
            case InvoiceAction::TYPE_PAYOUT:
                $this->invoice->setPaidOut(true);
                break;
            case InvoiceAction::TYPE_PAYDIRECT:
                $this->invoice->setPaymentReceived(true);
                $this->invoice->setPaidOut(true);
                break;
            case InvoiceAction::TYPE_NOTIFIED_OVERDUE:
            case InvoiceAction::TYPE_NOTIFIED_IN_DEFAULT:
                // this does not change any state
                break;
            default:
                throw new Exception("Invalid action type.");
        }
    }

}