<?php

namespace ProgrammaBill\Entities;

use Exception;
use Doctrine\ORM\Mapping AS ORM;
use ProgrammaBill\Interfaces\APIObject;

/**
 * An InvoiceLink is a link from an Invoice to a remote entity,
 * such as a backup copy or the invoice's representation in an external
 * billing or bookkeeping system.
 *
 * @ORM\Entity
 * @ORM\Table(name="mi_invoice_links")
 */
class InvoiceLink extends BaseEntity implements APIObject {
 
    /** @ORM\ManyToOne(targetEntity="Invoice") **/
    private $invoice;

    /** @ORM\ManyToOne(targetEntity="RemoteService") **/
    private $service;
    
    /** @ORM\ManyToOne(targetEntity="Actor") **/
    private $actor;

    /** @ORM\Column(type="string", length=50) **/
    private $ext_document_id;

    /** @ORM\Column(type="json_array") **/
    private $meta = [];

    public function __construct(Invoice $invoice, RemoteService $service, Actor $actor = null,
            string $ext_document_id = null, array $meta = []) {
        
        parent::__construct();
        $this->invoice = $invoice;
        $this->service = $service;
        $this->actor = $actor;
        $this->ext_document_id = $ext_document_id;
        $this->meta = $meta;
    }

    /**
     * Get the invoice this link is for.
     */
    public function getInvoice() : Invoice {
        return $this->invoice;
    }

    /**
     * Get the actor on whose behalf this link was established,
     * or NULL if link was created on behalf of the marketplace.
     */
    public function getActor() : ?Actor {
        return $this->actor;
    }

    /**
     * Indicates the actor type, either "merchant", "customer" or "marketplace".
     */
    public function getActorType() : string {
        $actor = $this->getActor();
        if (!isset($actor))
            return 'marketplace';
        
        $transaction = $this->getInvoice()->getTransaction();
        if ($actor == $transaction->getCustomer())
            return 'customer';
        if ($actor == $transaction->getMerchant())
            return 'merchant';
        
        return null;
    }

    /**
     * Get the remote service to which this link connects.
     */
    public function getService() : RemoteService {
        return $this->service;
    }

    /**
     * Get the remote identifier for the invoice.
     */
    public function getExternalDocumentID() : ?string {
        return $this->ext_document_id;
    }

    /**
     * Set the remote identifier for the invoice.
     * 
     * @param string $ext_document_id
     * @return InvoiceLink self-reference for chaining
     */
    public function setExternalDocumentID(string $ext_document_id) : InvoiceLink {
        $this->ext_document_id = $ext_document_id;
        return $this;
    }

    /**
     * Get user-supplied meta-data for this link.
     * @return array
     */
    public function getMeta() : array {
        return $this->meta;
    }

    /**
     * Set user-supplied meta-data for this link.
     *
     * @param array $user_meta
     * @return InvoiceLink self-reference for chaining
     */
    public function setMeta(array $meta) : InvoiceLink {
        $this->meta = $meta;
        return $this;
    }

    public function parseAPIRequest(Actor $actor = null, array $data, array $options = []) {
        throw new Exception("Not used.");
    }

    public static function isValidAPIRequest(array $data, array $options = []) {
        throw new Exception("Not used.");
    }

    public function createAPIResponse(Actor $actor = null, array $options = []) {
        return [
            'actor' => $this->getActorType(),
            'created_timestamp' => $this->getCreatedTimestamp()->format('c'),
            'service' => (string)$this->getService()->getCOID(),
            'ext_document_id' => $this->getExternalDocumentID(),
            'meta' => $this->getMeta()
        ];
    }

}