<?php

namespace ProgrammaBill\Entities;

use Exception;
use Doctrine\ORM\Mapping AS ORM;
use Doctrine\Common\Collections\ArrayCollection;
use Carbon\Carbon;
use Punic\Territory;
use ML\IRI\IRI;
use CloudObjects\SDK\ObjectRetriever,
    CloudObjects\SDK\NodeReader;
use ProgrammaBill\Interfaces\APIObject,
    ProgrammaBill\Interfaces\Address;

/**
 * A Marketplace is a CloudObjects service which functions as a platform on which
 * accounts can have transactions with each other.
 *
 * @ORM\Entity(repositoryClass="ProgrammaBill\Repositories\MarketplaceRepository")
 * @ORM\Table(name="mi_marketplaces")
 */
class Marketplace extends BaseEntity implements APIObject, Address {

    /** @ORM\Column(type="string", length=50, unique=true) **/
    private $service_domain;

    /**
     * @ORM\OneToMany(targetEntity="Transaction", mappedBy="marketplace")
     * @ORM\OrderBy({ "id" = "desc" })
     **/
    private $transactions;

    private $coid;
    private $objectRetriever;
    private $object;
    private $reader;

    private function getReader() {
        if (!isset($this->reader))
            $this->reader = new NodeReader([
                'prefixes' => [
                    'mi' => 'coid://marketplace-invoicing.cloudobjects.io/'
                ]
            ]);
    
        return $this->reader;
    }

    /**
     * Creates a new marketplace.
     * @param string $service_domain CloudObjects service namespace
     */
    public function __construct(string $service_domain) {
        parent::__construct();
        $this->service_domain = $service_domain;
        $this->transactions = new ArrayCollection;
    }

    /**
     * Get the CloudObjects service namespace for this marketplace.
     * @return string
     */
    public function getServiceDomain() {
        return $this->service_domain;
    }

    /**
     * Get the CloudObjects object representing the marketplace.
     */
    public function getObject() {
        if (!isset($this->objectRetriever))
            throw new Exception("Must assign ObjectRetriever before retrieving object.");

        if (!isset($this->object))
            $this->object = $this->objectRetriever->getObject($this->getCOID());

        return $this->object;
    }

    /**
     * Get the COID for the marketplace service.
     */
    public function getCOID() {
        if (!isset($this->coid))
            $this->coid = new IRI('coid://'.$this->getServiceDomain());

        return $this->coid;
    }

    /**
     * Get all transactions for the marketplace.
     */
    public function getTransactions() {
        return $this->transactions;
    }

    public function parseAPIRequest(Actor $actor = null, array $data, array $options = []) {
        throw new Exception("Not used.");
    }

    public static function isValidAPIRequest(array $data, array $options = []) {
        throw new Exception("Not used.");
    }

    public function createAPIResponse(Actor $actor = null, array $options = []) {    
        $r = $this->getReader();
        $o = $this->getObject();
        return [
            'org_name' => $r->getFirstValueString($o, 'mi:hasOrgName'),
            'dep_name' => $r->getFirstValueString($o, 'mi:hasDepName'),
            'full_name' => $r->getFirstValueString($o, 'mi:hasFullName'),
            'street_address' => $r->getFirstValueString($o, 'mi:hasStreetAddress'),
            'postal_code' => $r->getFirstValueString($o, 'mi:hasPostalCode'),
            'location' => $r->getFirstValueString($o, 'mi:hasLocation'),
            'country' => $this->getCountry(),
            'company_tax_id' => $r->getFirstValueString($o, 'mi:hasCompanyTaxID'),
            'cost_center_id' => $r->getFirstValueString($o, 'mi:hasCostCenterID'),
            'eu_vat_id' => $r->getFirstValueString($o, 'mi:hasEUVATID'),
            'vat_exempt' => $this->isVATExempt(),
            'registration_info_line' => $r->getFirstValueString($o, 'mi:hasRegistrationInfoLine'),
            'management_info_line' => $r->getFirstValueString($o, 'mi:hasManagementInfoLine')
        ];
    }

    /**
     * Get the cycle that is used for invoice numbering for invoices
     * sent from this marketplace. One of "Y", "m" or "d".
     * @return string
     */
    public function getInvoicingCycle() {
        // TODO: decide whether to store this in DB or load from CO
        return "Y";
    }

    /**
     * Get the desired number for the very first invoice, if there's any.
     * @return int|null
     */
    public function getFirstInvoiceNumber() {
        return $this->getReader()->getFirstValueInt($this->getObject(), 'mi:hasFirstInvoiceNumber');
    }

    /**
     * Get the name of the S3 bucket in which invoice documents for
     * this marketplaces are stored.
     */
    public function getBucketName() {
        return "co-mi-".str_replace(".", "--", $this->getServiceDomain());
    }

     /**
     * Get the country in which the actor is located
     * as an ISO-3166-1 code.
     * 
     * @return string|null
     */
    public function getCountry() {
        return $this->getReader()->getFirstValueString($this->getObject(), 'mi:hasCountry');
    }

    /**
     * Indicates whether the actor is located in the European Union.
     * @return bool
     */
    public function isLocatedInEU() : bool {
        return in_array($this->getCountry(), Territory::getChildTerritoryCodes('EU'));
    }

    /**
     * Indicates whether this actor is exempt from charging VAT.
     */ 
    public function isVATExempt()
    {
        return false; // fair assumption that marketplaces are never VAT exempt
    }

    /**
     * Generate a due date for an invoice.
     * 
     * @param Carbon $creationDate The date on which the invoice was created.
     * @return Carbon
     */
    public function getDueDateFromCreationDate(Carbon $creationDate) {
        // TODO: decide whether to store this in DB or load from CO
        return $creationDate->copy()->addDays(7);
    }

    /**
     * Marketplaces are always businesses, thus this always returns "true".
     * 
     * @return bool
     */
    public function isBusiness() {
        return true;
    }

    public function setObjectRetriever(ObjectRetriever $objectRetriever) {
        $this->objectRetriever = $objectRetriever;
        return $this;
    }

    public function getObjectRetriever() {
        return $this->objectRetriever;
    }

    public function getTemplateContext() {
        $context = [];

        if ($this->getReader()->hasProperty($this->getObject(), 'mi:hasAttachedLogoFile')) {
            // Fetch attached logo, and add as Base64
            $filename = $this->getReader()->getFirstValueString($this->getObject(), 'mi:hasAttachedLogoFile');
            $contentType = 'application/octet-stream';
            switch (substr($filename, strrpos($filename, '.'))) {
                case ".ico":
                    $contentType = 'image/x-icon';
                    break;
                case ".png":
                    $contentType = 'image/png';
                    break;
                case ".svg":
                    $contentType = 'image/svg+xml';
                    break;
                case ".jpg":
                case ".jpeg":
                    $contentType = 'image/jpeg';
                    break;
            }

            $logoContent = $this->objectRetriever->getAttachment($this->getCOID(), $filename);
            $context['logo_src'] = 'data:'.$contentType.';base64,'.base64_encode($logoContent);
        }

        return $context;
    }

    /**
     * Get a list of services for autolinking actors into one identity based on matching email.
     */
    public function getAutolinkServiceCOIDsForMatchingEMail() {
        return $this->getReader()->getAllValuesIRI($this->getObject(), 'mi:autolinksWithMatchingEMailActorsOn');
    }

    /**
     * Get a delay in days that should be applied before notifying the service of an overdue invoice.
     */
    public function getOverdueDelay() {
        return 1; // currently hardcoded to 1 day, can be a configuration option later
    }

    /**
     * Get a timespan in days after which an overdue invoice should be considered in default
     * after sending the first overdue reminder.
     */
    public function getInDefaultAfter() {
        return 7; // currently hardcoded to 7 days, can be a configuration option later
    }

}
