<?php

namespace ProgrammaBill\Entities;

use DateTime, Exception;
use Doctrine\ORM\Mapping AS ORM;
use Doctrine\Common\Collections\ArrayCollection;
use Webmozart\Assert\Assert;

use ProgrammaBill\Interfaces\APIObject;
use ProgrammaBill\Helpers;

/**
 * A Transaction is an exchange of goods or services on the marketplace for which
 * multiple invoices can be generated and associated.
 *
 * @ORM\Entity(repositoryClass="ProgrammaBill\Repositories\TransactionRepository")
 * @ORM\Table(name="mi_transactions")
 */
class Transaction extends BaseEntity implements APIObject {

    /** @ORM\ManyToOne(targetEntity="Marketplace") **/
    private $marketplace;

    /** @ORM\ManyToOne(targetEntity="Actor") **/
    private $merchant;

    /** @ORM\ManyToOne(targetEntity="Actor") **/
    private $customer;

    /** @ORM\ManyToOne(targetEntity="ActorDetail") **/
    private $merchant_detail;

    /** @ORM\ManyToOne(targetEntity="ActorDetail") **/
    private $customer_detail;

    /** @ORM\Column(type="string", length=50, nullable=true) **/
    private $ext_contract_id;

    /** @ORM\Column(type="json_array") **/
    private $meta = [];

    /** @ORM\Column(type="datetime", nullable=true) **/
    private $delivery_date;

    /** @ORM\Column(type="datetime", nullable=true) **/
    private $delivery_date_end;

    /** @ORM\OneToMany(targetEntity="Invoice", mappedBy="transaction", cascade={"persist"}) **/
    private $invoices;

    /** @ORM\OneToOne(targetEntity="Transaction") **/
    private $superseded_transaction;

    /** @ORM\OneToOne(targetEntity="Transaction", mappedBy="superseded_transaction") **/
    private $superseding_transaction;

    /** @ORM\ManyToOne(targetEntity="Transaction") **/
    private $parent_transaction;

    /** @ORM\OneToMany(targetEntity="Transaction", mappedBy="parent_transaction") **/
    private $child_transactions;

    // temporary helper fields

    private $invoices_for_actor = [];
    private $invoices_for_platform = null;

    /**
     * Creates a new transaction.
     * @param Marketplace $marketplace The platform on which this transaction takes place.
     */
    public function __construct(Marketplace $marketplace) {
        parent::__construct();
        $this->marketplace = $marketplace;

        $this->invoices = new ArrayCollection;
        $this->child_transactions = new ArrayCollection;
    }

    /**
     * Get the platform on which this transaction takes place.
     * @return Marketplace
     */
    public function getMarketplace() {
        return $this->marketplace;
    }

    /**
     * Get the actor who is the merchant in this transaction, if any.
     * @return Actor|null
     */
    public function getMerchant() {
        return $this->merchant;
    }

    /**
     * Set the actor who is the merchant in this transaction.
     * @param Actor $merchant
     */
    public function setMerchant(Actor $merchant) {
        $this->merchant = $merchant;
        return $this;
    }

    /**
     * Get the actor who is the consumer in this transaction, if any.
     */
    public function getCustomer() {
        return $this->customer;
    }

    /**
     * Set the actor who is the consumer in this transaction.
     * @param Actor $customer
     */
    public function setCustomer(Actor $customer) {
        $this->customer = $customer;
        return $this;
    }

    /**
     * Get the details for the actor who is the consumer in this transaction, if any.
     */
    public function getCustomerDetail() : ?ActorDetail {
        return $this->customer_detail;
    }

    /**
     * Set the actor detail for the consumer in this transaction.
     * Fails if customer wasn't set or doesn't match details.
     * 
     * @param ActorDetail $customer_detail
     */
    public function setCustomerDetail(ActorDetail $customer_detail) : Transaction {
        Assert::eq($customer_detail->getActor(), $this->getCustomer(),
            "Customer details must belong to assigned customer.");
        $this->customer_detail = $customer_detail;
        return $this;
    }

    /**
     * Get the details for the actor who is the merchant in this transaction, if any.
     */
    public function getMerchantDetail() : ?ActorDetail {
        return $this->merchant_detail;
    }

    /**
     * Set the actor detail for the merchant in this transaction.
     * Fails if merchant wasn't set or doesn't match details.
     * 
     * @param ActorDetail $merchant_detail
     */
    public function setMerchantDetail(ActorDetail $merchant_detail) : Transaction {
        Assert::eq($merchant_detail->getActor(), $this->getMerchant(),
            "Merchant details must belong to assigned merchant.");
        $this->merchant_detail = $merchant_detail;
        return $this;
    }

    /**
     * Get the ID of the contract that this transaction relates to, if any.
     */
    public function getContractID() {
        return $this->ext_contract_id;
    }

    /**
     * Set the ID of the contract that this transaction relates to, as
     * an ID from an external system.
     * 
     * @param string $ext_contract_id
     * @return Transaction self-reference for chaining
     */
    public function setExternalContractID(string $ext_contract_id) {
        $this->ext_contract_id = $ext_contract_id;
        return $this;
    }

    /**
     * Get user-supplied meta-data for this transaction.
     * @return array
     */
    public function getMeta() {
        return $this->meta;
    }

    /**
     * Set user-supplied meta-data for this transaction.
     *
     * @param array $user_meta
     * @return Transaction self-reference for chaining
     */
    public function setMeta(array $meta) {
        $this->meta = $meta;
        return $this;
    }

    /**
     * Get all invoices associated with this transaction.
     * 
     * @return ArrayCollection
     */
    public function getInvoices() {
        return $this->invoices;
    }

    /**
     * Get the invoice that matches the provided type,
     * if any exists in the transaction, null otherwise.
     * 
     * @param int $invoice_type
     * @return Invoice|null
     */
    public function getInvoiceWithType($invoice_type) {
        foreach ($this->getInvoices() as $i)
            if ($i->getInvoiceType() == $invoice_type)
                return $i;

        return null;
    }

    /**
     * Get the invoice that matches the provided type,
     * if any exists in the transaction, null otherwise.
     * 
     * @param int $invoice_type
     * @return Invoice|null
     */
    public function getInvoiceWithTypeAsString($invoice_type) {
        foreach ($this->getInvoices() as $i)
            if ($i->getInvoiceTypeAsString() == $invoice_type)
                return $i;

        return null;
    }    

    /**
     * Get the invoice that matches the provided number,
     * if it exists in the transaction, null otherwise.
     * 
     * @param string $number
     * @return Invoice|null
     */
    public function getInvoiceWithNumber($number) {
        foreach ($this->getInvoices() as $i)
            if ($i->getNumber() == $number)
                return $i;

        return null;
    }

    /**
     * Get all invoices associated with this transactions that are relevant for
     * and visible to the specified actor.
     * 
     * @param Actor $actor
     * @return array
     */
    public function getInvoicesForActor(Actor $actor) {
        if (!isset($this->invoices_for_actor[$actor->getId()])) {
            $this->invoices_for_actor[$actor->getId()] = [];
      
            foreach ($this->getInvoices() as $i) {
                if (in_array($i->getInvoiceType(), [ Invoice::TYPE_MERCHANT_TO_CUSTOMER, Invoice::TYPE_MERCHANT_TO_CUSTOMER_REVERSED ])
                    || ($actor == $this->getMerchant() && in_array($i->getInvoiceType(),
                        [ Invoice::TYPE_PLATFORM_TO_MERCHANT, Invoice::TYPE_PLATFORM_TO_MERCHANT_REVERSED,
                            Invoice::TYPE_MERCHANT_TO_PLATFORM, Invoice::TYPE_MERCHANT_TO_PLATFORM ] ))
                    || ($actor == $this->getCustomer() && in_array($i->getInvoiceType(),
                        [ Invoice::TYPE_PLATFORM_TO_CUSTOMER, Invoice::TYPE_PLATFORM_TO_CUSTOMER_REVERSED ]))
                ) $this->invoices_for_actor[$actor->getId()][] = $i;
            }
        }

        return $this->invoices_for_actor[$actor->getId()];
    }

    /**
     * Get all invoices from or to the platform.
     */
    public function getInvoicesForPlatform() {
        if (!isset($this->invoices_for_platform)) {
            $this->invoices_for_platform = [];
      
            foreach ($this->getInvoices() as $i) {
                if ($i->getSenderType() == Invoice::ACTOR_PLATFORM || $i->getReceiverType() == Invoice::ACTOR_PLATFORM) {
                    $this->invoices_for_platform[] = $i;
                }
            }
        }

        return $this->invoices_for_platform;
    }

    /**
     * Get the due date of the earliest due invoice for the specified actor.
     * 
     * @param Actor $actor
     * @return DateTime
     */
    public function getDueDate(Actor $actor = null) {
        $date = null;
        $invoices = isset($actor)
            ? $this->getInvoicesForActor($actor)
            : $this->getInvoices();
        foreach ($invoices as $i)
        if (!isset($date) || ($date > $i->getDueDate()))
            $date = $i->getDueDate();
        return $date;
    }

    /**
     * Indicates whether this is an outbound transaction from the perspective of
     * a certain actor i.e. if that actor is the merchant.
     *
     * @param Actor $actor
     * @return boolean
     */
    public function isOutbound(Actor $actor = null) {
        return ($actor == $this->getMerchant());
    }

    /**
     * Indicates whether payment has been received for all invoices in this
     * transactions that are relevant for the specified actor.
     * 
     * @param Actor $actor
     * @return bool
     */
    public function isPaymentReceived(Actor $actor) {
        foreach ($this->getInvoicesForActor($actor) as $i)
            if (!$i->isPaymentReceived())
                return false;
        
        return true;
    }

    /**
     * Indicates whether all merchant invoices in this
     * transaction have been paid out to the merchant.
     * 
     * @return bool
     */
    public function isPaidOut() {
        foreach ($this->getInvoices() as $i)
            if ($i->getInvoiceType() == Invoice::TYPE_MERCHANT_TO_CUSTOMER && !$i->isPaidOut())
                return false;
        
        return true;
    }

    /**
     * Indicates whether this transaction has been reversed.
     * 
     * @return bool
     */
    public function isReversed() {
        foreach ($this->getInvoices() as $i)
            if ($i->getReversedOriginalNumber() != null)
                return true;

        return false;
    }

    public function parseAPIRequest(Actor $actor = null, array $data, array $options = []) {
        if (isset($actor)) {
            // For requests with actors, make sure that the actor account
            // is used as merchant or customer for this transaction
            if ($data['actor'] == 'merchant')
                $this->setMerchant($actor);
            elseif ($data['actor'] == 'customer')
                $this->setCustomer($actor);
        }

        if (isset($options['_actor_repository'])) {
            if ($this->getMerchant() == null && isset($data['merchant'])) {
                // We have no merchant so we need to find it; look up by external user ID if available, email otherwise
                if (isset($data['merchant']['ext_user_id']) && !empty($data['merchant']['ext_user_id']))
                    $this->setMerchant($options['_actor_repository']
                        ->fetchOrCreateWithExternalUserID($data['merchant']['ext_user_id'], $this->getMarketplace(), @$data['merchant']['email']));
                elseif (isset($data['merchant']['email']) && !empty($data['merchant']['email']))
                    $this->setMerchant($options['_actor_repository']
                        ->fetchOrCreateWithEmail($data['merchant']['email'], $this->getMarketplace()));
            }
            if ($this->getCustomer() == null && isset($data['customer'])) {
                // We have no customer so we need to find it; look up by external user ID if available, email otherwise
                if (isset($data['customer']['ext_user_id']) && !empty($data['customer']['ext_user_id']))
                    $this->setCustomer($options['_actor_repository']
                        ->fetchOrCreateWithExternalUserID($data['customer']['ext_user_id'], $this->getMarketplace(), @$data['customer']['email']));
                elseif (isset($data['customer']['email']) && !empty($data['customer']['email']))
                    $this->setCustomer($options['_actor_repository']
                        ->fetchOrCreateWithEmail($data['customer']['email'], $this->getMarketplace()));
            }
        } 

        if ($this->getMerchant() != null) {
            $this->getMerchant()->parseAPIRequest($actor, $data['merchant']);
            $this->setMerchantDetail($this->getMerchant()->getOrCreateCurrentDetail());
        }
        if ($this->getCustomer() != null) {
            $this->getCustomer()->parseAPIRequest($actor, $data['customer']);
            $this->setCustomerDetail($this->getCustomer()->getOrCreateCurrentDetail());
        }

        if (isset($actor) && $actor->getId() == null
                && isset($options['_actor_repository'])) {
            // Our actor is brand new, let's see if we can link
            // them to an existing identity
            $options['_actor_repository']->linkWithExistingAccounts($actor);
        }

        if (isset($data['delivery_date']))
            $this->setDeliveryDate(new DateTime($data['delivery_date']));
    
        if (isset($data['delivery_date_end']))
            $this->setDeliveryDateEnd(new DateTime($data['delivery_date_end']));

        if (isset($data['contract_id']))
            $this->setExternalContractID($data['contract_id']);

        if (isset($options['_transaction_repository']) && isset($data['supersedes_transaction_id'])) {
            // Create a superseding transaction
            $oldTransaction = $options['_transaction_repository']->findOneById($data['supersedes_transaction_id']);
            Assert::true(isset($oldTransaction) && $oldTransaction->getMarketplace() == $this->getMarketplace(),
                "Superseded transaction not found.");
            Assert::eq($oldTransaction->getMerchant(), $this->getMerchant(),
                "Superseded transaction must have same merchant.");
            Assert::eq($oldTransaction->getCustomer(), $this->getCustomer(),
                "Superseded transaction must have same customer.");

            $this->setSupersededTransaction($oldTransaction);
        }

        if (isset($options['_transaction_repository']) && isset($data['parent_transaction_id'])) {
            // Create a child transaction
            $parentTransaction = $options['_transaction_repository']->findOneById($data['parent_transaction_id']);
            Assert::true(isset($parentTransaction) && $parentTransaction->getMarketplace() == $this->getMarketplace(),
                "Parent transaction not found.");

            // TODO: define rules for parent transaction

            $this->setParentTransaction($parentTransaction);
        }

        foreach ($data['invoices'] as $i) {
            if (isset($i['number'])) {
                // Update existing invoice
                foreach ($this->getInvoices() as $invoice) {
                    if ($invoice->getNumber() == $i['number'])
                        $invoice->parseAPIRequest($actor, $i, $options);
                }
            } else
            if (isset($options['_invoice_repository'])) {
                // Create new invoice

                // Before creating new invoices in a superseding transaction,
                // make sure the old one was reversed
                if (isset($oldTransaction) && $oldTransaction->getInvoiceWithTypeAsString($i['type']) != null) {
                    Assert::notNull($oldTransaction->getInvoiceWithTypeAsString($i['type'])->getReverseInvoice(),
                        "Superseding transaction can only include invoice types that were reversed in the superseded transaction.");
                }
                
                switch ($i['type']) {
                    case "M2C":
                        $invoice = $options['_invoice_repository']
                            ->createNewInvoiceForMerchant($this, Invoice::TYPE_MERCHANT_TO_CUSTOMER);
                        break;
                    case "M2P":
                        $invoice = $options['_invoice_repository']
                            ->createNewInvoiceForMerchant($this, Invoice::TYPE_MERCHANT_TO_PLATFORM);
                        break;
                    case "P2M":
                        $invoice = $options['_invoice_repository']
                            ->createNewInvoiceForMarketplace($this, Invoice::TYPE_PLATFORM_TO_MERCHANT);
                        break;
                    case "P2C":
                        $invoice = $options['_invoice_repository']
                            ->createNewInvoiceForMarketplace($this, Invoice::TYPE_PLATFORM_TO_CUSTOMER);
                        break;
                    default:
                        throw new Exception("Invalid invoice type.");
                }
                $invoice->parseAPIRequest($actor, $i, $options);
                $this->getInvoices()->add($invoice);
            }
        }

        if (isset($data['meta']))
            $this->setMeta($data['meta']);
    }

    public static function isValidAPIRequest(array $data, array $options = []) {
        if (!Helpers::containsAllFields($data, [ 'invoices' ])
                || !is_array($data['invoices']))
            return false;

        if (!isset($options['_actorless']))
            if (!Helpers::containsAllFields($data, [ 'customer', 'merchant', 'actor' ])
                || !in_array($data['actor'], [ 'merchant', 'customer' ])
                || ($data['actor'] == 'customer' && (!is_array($data['customer']) || !isset($data['customer']['email'])))
                || ($data['actor'] == 'merchant' && (!is_array($data['merchant']) || !isset($data['merchant']['email'])))
            )
            return false;

        foreach ($data['invoices'] as $i)
            if (!Invoice::isValidAPIRequest($i, $options))
                return false;

        return true;
    }

    public function createAPIResponse(Actor $actor = null, array $options = []) {
        $options = array_merge([
            'list_invoices_for_actor_only' => isset($actor) ? true : false,
            'list_invoices' => true,
            'expand_linked_transactions' => false
        ], $options);

        $now = new DateTime;
        $response = [
            'id' => $this->getId(),
            'created_timestamp' => $this->getCreatedTimestamp()->format('c'),
            'contract_id' => $this->getContractID(),
            'merchant' => ($this->getMerchant() != null)
                ? $this->getMerchant()->createAPIResponse($actor,
                        array_merge($options, [ '_detail' => $this->getMerchantDetail() ])
                    )
                : null,
            'customer' => ($this->getCustomer() != null)
                ? $this->getCustomer()->createAPIResponse($actor,
                        array_merge($options, [ '_detail' => $this->getCustomerDetail() ])
                    )
                : null,
            'delivery_date' => ($this->getDeliveryDate() != null) ? $this->getDeliveryDate()->format('Y-m-d') : null,
            'delivery_date_end' => ($this->getDeliveryDateEnd() != null ? $this->getDeliveryDateEnd()->format('Y-m-d') : null),
            'delivery_date_is_period' => ($this->getDeliveryDateEnd() != null),
            'due_date' => $this->getDueDate($actor) ? $this->getDueDate($actor)->format('Y-m-d') : null,
            'stage' => $this->isReversed()
                ? 'reversed'
                : (isset($actor)
                    ? ($this->isOutbound($actor)
                        ? ($this->isPaidOut()
                            ? 'paid_out'
                            : ($this->isPaymentReceived($actor) ? 'paid_in' : ($now > $this->getDueDate($actor) ? 'overdue' : 'due'))
                            )
                        : ($this->isPaymentReceived($actor)
                            ? 'paid'
                            : ($now > $this->getDueDate($actor) ? 'overdue' : 'due')
                            )
                        )
                    : null),
            'meta' => $this->getMeta(),
        ];
        
        if ($this->getSupersededTransaction() != null) {
            if ($options['expand_linked_transactions'])
                $response['supersedes_transaction'] = $this->getSupersededTransaction()->createAPIResponse($actor,
                    array_merge($options, [ 'expand_linked_transactions' => false ]));
            else
                $response['supersedes_transaction_id'] = $this->getSupersededTransaction()->getId();
        }  
        if ($this->getSupersedingTransaction() != null) {
            if ($options['expand_linked_transactions'])
                $response['superseded_by_transaction'] = $this->getSupersedingTransaction()->createAPIResponse($actor,
                    array_merge($options, [ 'expand_linked_transactions' => false ]));
            else
                $response['superseded_by_transaction_id'] = $this->getSupersedingTransaction()->getId();
        }
        if ($this->getParentTransaction() != null) {
            if ($options['expand_linked_transactions'])
                $response['parent_transaction'] = $this->getParentTransaction()->createAPIResponse($actor,
                    array_merge($options, [ 'expand_linked_transactions' => false ]));
            else
                $response['parent_transaction_id'] = $this->getParentTransaction()->getId();
        }
        if (count($this->getChildTransactions()) > 0) {
            if ($options['expand_linked_transactions']) {
                $response['child_transactions'] = [];
                foreach ($this->getChildTransactions() as $t)
                    $response['child_transactions'][] = $t->createAPIResponse($actor,
                    array_merge($options, [ 'expand_linked_transactions' => false ]));
            } else {
                $response['child_transactions_ids'] = [];
                foreach ($this->getChildTransactions() as $t)
                    $response['child_transaction_ids'][] = $t->getId();
            }
        }

        $sum = 0;

        if ($options['list_invoices']) {
            $response['invoices'] = [];

            $invoices = $options['list_invoices_for_actor_only']
                ? $this->getInvoicesForActor($actor)
                : $this->getInvoices();
            
            foreach ($invoices as $i) {
                $response['invoices'][] = $i->createAPIResponse($actor, $options);
                if ($i->isOutbound($actor)) {
                    $sum += $i->getTotalPayableAmount();
                } else {
                    $sum -= $i->getTotalPayableAmount();
                }
            }

            $response['currency'] = $i->getCurrency();
            $response['effective_sum'] = abs($sum); // TODO: handle different currencies
        }        

        return $response;
    }

    /**
     * Get the date on which the product or service was delivered. If the transaction
     * covers multiple days, returns the first date.
     * 
     * @return DateTime
     */ 
    public function getDeliveryDate() {
        return $this->delivery_date;
    }

    /**
     * Set the date on which the product or service was delivered. If the transaction
     * covers multiple days, set the first date.
     * 
     * @param DateTime $delivery_date
     * @return self
     */ 
    public function setDeliveryDate(DateTime $delivery_date) {
        $this->delivery_date = $delivery_date;
        return $this;
    }

    /**
     * If the transaction covers multiple days, get the last day on which a product
     * or service was delivered. Returns null otherwise.
     * 
     * @return DateTime|null
     */ 
    public function getDeliveryDateEnd() {
        return $this->delivery_date_end;
    }

    /**
     * If the transaction covers multiple days, set the last day on which a product
     * or service was delivered.
     * 
     * @param DateTime $delivery_date_end
     * @return self
     */ 
    public function setDeliveryDateEnd(DateTime $delivery_date_end) {
        $this->delivery_date_end = $delivery_date_end;
        return $this;
    }

    /**
     * If this transaction supersedes another transaction, retrieve the original transaction.
     * 
     * @return Transaction|null
     */
    public function getSupersededTransaction() {
        return $this->superseded_transaction;
    }

    /**
     * Specifies another transaction that this transaction should supersede.
     * 
     * @param Transaction $superseded_transaction
     * @return self
     */
    public function setSupersededTransaction(Transaction $superseded_transaction) {
        $this->superseded_transaction = $superseded_transaction;
        return $this;
    }

    /**
     * If this transaction was superseded by another transaction, retrieve the newer transaction.
     * 
     * @return Transaction|null
     */
    public function getSupersedingTransaction() {
        return $this->superseding_transaction;
    }

    /**
     * If this transaction has a parent, retrieve parent transaction.
     * 
     * @return Transaction|null
     */
    public function getParentTransaction() {
        return $this->parent_transaction;
    }

    /**
     * Specifies another transaction that this transaction should become a child of.
     * 
     * @param Transaction $parent_transaction
     * @return self
     */
    public function setParentTransaction(Transaction $parent_transaction) {
        $this->parent_transaction = $parent_transaction;
        return $this;
    }

    /**
     * Get the list of child transactions.
     * 
     * @return ArrayCollection
     */
    public function getChildTransactions() {
        return $this->child_transactions;
    }

}
