<?php

namespace ProgrammaBill;

use Exception;
use Psr\Container\ContainerInterface;
use Aws\S3\S3Client;
use ML\JsonLD\Node;
use Michelf\MarkdownExtra;
use Mpdf\Mpdf;
use CloudObjects\SDK\ObjectRetriever, CloudObjects\SDK\NodeReader;
use CloudObjects\SDK\Helpers\SDKLoader;
use ProgrammaBill\Entities\Actor,
    ProgrammaBill\Entities\Invoice;

class InvoiceFormatter {

    private $twig;
    private $s3;
    private $retriever;

    public function __construct(ContainerInterface $container, ObjectRetriever $retriever,
            SDKLoader $sdkLoader) {
        
        $this->twig = $container->get('twig');
        $this->twig->getExtension(\Twig\Extension\CoreExtension::class)
            ->setNumberFormat(2, ',', '');
        $this->twig->getExtension(PunicTwig::class)->setLocale('de_DE');
        
        $this->retriever = $retriever;
        $this->s3 = $sdkLoader->get(S3Client::class, [
            'version' => 'latest',
            'region' => 'eu-central-1'
        ]);
        $this->reader = new NodeReader([
            'prefixes' => [ 'mi' => 'coid://marketplace-invoicing.cloudobjects.io/' ]
        ]);
    }

    private function generateSimpleTextPaymentInstruction(Invoice $invoice, Node $instruction,
            array $languages, array $data) {
        
        $texts = [];
        foreach ($languages as $lang) {
            $template = $this->reader->getLocalizedString($instruction,
                'mi:containsFooterTextTemplate', $lang);
            $template = substr(MarkdownExtra::defaultTransform($template), 3, -5);
            $data['merchant'] = $invoice->getTransaction()->getMerchant();
            $data['customer'] = $invoice->getTransaction()->getCustomer();
            $texts[$lang] = $this->twig->createTemplate($template)->render($data);
        }
        return $texts;
    }

    private function generatePaymentInstruction(Invoice $invoice, array $languages, array $data) {
        if ($invoice->getPaymentInstruction() == null)
            return null;
        
        $instruction = $this->retriever->get($invoice->getPaymentInstruction()->getCOID());
        if (!isset($instruction))
            throw new Exception("Could not retrieve payment instruction details.");
        
        if ($this->reader->hasType($instruction, 'mi:SimpleTextPaymentInstructionType'))
            return $this->generateSimpleTextPaymentInstruction($invoice, $instruction,
                $languages, $data);
        
        throw new Exception("Unsupported payment instruction type.");
    }

    public function getHTML(Invoice $invoice, Actor $actor = null) {
        try {
            // Poke to see if we can get the object
            $invoice->getMarketplace()->getObject();
        } catch (Exception $e) {
            // Potential cause for exception is that we're unable to access the
            // marketplace object due to retriever non-assignment, which we can
            // fix now and retry

            $invoice->getMarketplace()->setObjectRetriever($this->retriever);
        }

        $data = $invoice->createAPIResponse($actor, [
            'embed_items' => true,
            'embed_tax_summary' => true,
            'embed_addresses' => true,
            'embed_transaction_dates_only' => true
        ]);
        $data['lang2'] = 'en';
        $data['platform'] = $invoice->getMarketplace()->getTemplateContext();
        
        $data['payment_instruction_rendered'] = $this->generatePaymentInstruction($invoice, [ 'de', 'en' ], $data);
        $data['display_vat_per_line'] = (!$data['sender']['vat_exempt'] && count($data['tax_summary']) > 1);
        $extra = [];
        if ($invoice->getTaxDisplay() != Invoice::TAX_DISPLAY_DEFAULT) {
            // We have a special tax display type, so we overwrite the B2B/B2C template selection
            switch ($invoice->getTaxDisplay()) {
                case Invoice::TAX_DISPLAY_INCLUDE:
                    $extra['suffix'] = 'b2c';
                    break;
                case Invoice::TAX_DISPLAY_EXCLUDE:
                    $extra['suffix'] = 'b2b';
                    break;
            }
        }
        $template = TemplateFinder::getTemplateFilenameFor(
            TemplateFinder::INVOICE, $invoice->getSender(), $invoice->getReceiver(), $extra);
        return $this->twig->render($template, $data);
    }

    public function getPDF(Invoice $invoice, Actor $actor = null) {
        $mpdf = new Mpdf;
        $mpdf->WriteHTML($this->getHTML($invoice, $actor));
        return $mpdf->Output('', 'S');
    }

    private function getS3Parameters(Invoice $invoice) {
        return [
            'Bucket' => $invoice->getTransaction()->getMarketplace()->getBucketName(),
            'Key' => ($invoice->getSenderType() == Invoice::ACTOR_MERCHANT
                ? md5($invoice->getTransaction()->getMerchant()->getAAUID()).'/' : '')
            . $invoice->getNumber().'.pdf'
        ];
    }

    public function createAndUploadPDF(Invoice $invoice, Actor $actor = null) {
        $this->s3->putObject(array_merge($this->getS3Parameters($invoice), [
            'Body' => $this->getPDF($invoice, $actor)
        ]));
    }

    public function getSignedPDFURL(Invoice $invoice, $valid = '+30 minutes') {
        $command = $this->s3->getCommand('GetObject', $this->getS3Parameters($invoice));
        return (string)$this->s3->createPresignedRequest($command, $valid)->getUri();
    }

}
