<?php

namespace ProgrammaBill\MarketplaceAPI;

use Symfony\Bundle\FrameworkBundle\Controller\AbstractController;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException,
    Symfony\Component\HttpKernel\Exception\BadRequestHttpException;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Component\HttpFoundation\Request;
use Doctrine\ORM\EntityManagerInterface;
use ProgrammaBill\AuthenticationSubscriber, ProgrammaBill\InvoiceFormatter,
    ProgrammaBill\FilterQuerySet;
use ProgrammaBill\Entities\Marketplace, ProgrammaBill\Entities\Transaction,
    ProgrammaBill\Entities\Invoice, ProgrammaBill\Entities\InvoiceAction;

/**
 * @Route("/api")
 */
class Invoices extends AbstractController {

    private $as;
    private $repository;
    private $formatter;

    public function __construct(AuthenticationSubscriber $authenticationSubscriber,
            EntityManagerInterface $entityManager, InvoiceFormatter $formatter) {
        
        $this->as = $authenticationSubscriber;
        $this->repository = $entityManager->getRepository(Invoice::class);
        $this->formatter = $formatter;
    }

    /**
     * @Route("/platform-invoices", methods="GET")
     */
    public function getPlatformInvoices(Request $r) {
        $repository = $this->getDoctrine()->getRepository(Transaction::class);
               
        // Query repository
        $transactions = $repository->getAllForMarketplace(
            $this->as->getMarketplace(),
            FilterQuerySet::fromRequest($r, [ 'from', 'to' ])
        );

        $items = [];
        foreach ($transactions as $t) {
            foreach ($t->getInvoicesForPlatform() as $i) {
                $items[] = $i->createAPIResponse(null, [
                    '_invoice_formatter' => $this->formatter,
                    'embed_transaction' => true
                ]);
            }
        }

        return $this->json([
            'items' => $items
        ]);
    }

    /**
     * @Route("/merchants/{merchant_email}/invoices/{number}", methods="GET")
     */
    public function getInvoice($merchant_email, $number) {
        $invoice = $this->repository->getM2CWithMarketplaceMerchantEMailAndNumber(
            $this->as->getMarketplace(), $merchant_email, $number);

        if (!isset($invoice))
            throw new NotFoundHttpException;

        return $this->json($invoice->createAPIResponse($invoice->getTransaction()->getMerchant(), [
            'embed_items' => true,
            'embed_tax_summary' => true,
            'embed_addresses' => true,
            'embed_actions' => true,
            '_invoice_formatter' => $this->formatter
        ]));
    }

    /**
     * @Route("/merchants/{merchant_email}/invoices/{number}/actions", methods="GET")
     */
    public function getActions($merchant_email, $number) {
        $invoice = $this->repository->getM2CWithMarketplaceMerchantEMailAndNumber(
            $this->as->getMarketplace(), $merchant_email, $number);

        if (!isset($invoice))
            throw new NotFoundHttpException;
        
        $actions = [];
        foreach ($invoice->getActions() as $action)
            $actions[] = $action->createAPIResponse($invoice->getTransaction()->getMerchant());
        return $this->json([ 'actions' => $actions ]);
    }

    /**
     * @Route("/merchants/{merchant_email}/invoices/{number}/actions", methods="POST")
     */
    public function processAction($merchant_email, $number, Request $r) {
        $invoice = $this->repository->getM2CWithMarketplaceMerchantEMailAndNumber(
            $this->as->getMarketplace(), $merchant_email, $number);

        if (!isset($invoice))
            throw new NotFoundHttpException;

        // Read and validate input
        $data = json_decode($r->getContent(), true);
        if (!isset($data) || !InvoiceAction::isValidAPIRequest($data))
            throw new BadRequestHttpException;

        // We need an actor, but this API is called with service credentials without
        // reference to an account, so we simply decide to use the merchant
        $actor = $invoice->getTransaction()->getMerchant();

        // TODO: refactor and use what exists now in InvoiceAction class

        $action = new InvoiceAction($invoice);
        $action->parseAPIRequest($actor, $data);

        switch ($action->getActionType()) {
            case InvoiceAction::TYPE_PAYIN:
                if ($invoice->isPaymentReceived())
                    throw new BadRequestHttpException("Already received payment.");
                
                $invoice->setPaymentReceived(true);
                break;
            case InvoiceAction::TYPE_PAYOUT:
                if (!$invoice->isPaymentReceived())
                    throw new BadRequestHttpException("Invoice must have a payin before it can be market as paid out.");
                if ($invoice->isPaidOut())
                    throw new BadRequestHttpException("Already paid out.");
            
                $invoice->setPaidOut(true);
                break;
            case InvoiceAction::TYPE_PAYDIRECT:
                if ($invoice->isPaymentReceived())
                    throw new BadRequestHttpException("Already received payment.");
                if ($invoice->isPaidOut())
                    throw new BadRequestHttpException("Already paid out.");

                $invoice->setPaymentReceived(true);
                $invoice->setPaidOut(true);                
                break;
        }

        // Persist to database
        $invoice->getActions()->add($action);
        $this->getDoctrine()->getManager()->flush();

        $actions = [];
        foreach ($invoice->getActions() as $action)
            $actions[] = $action->createAPIResponse($actor);
        return $this->json([ 'actions' => $actions ]);
    }

}
