<?php

namespace ProgrammaBill\MarketplaceAPI;

use Symfony\Bundle\FrameworkBundle\Controller\AbstractController;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Component\HttpFoundation\Request,
    Symfony\Component\HttpFoundation\Response;
use Webmozart\Assert\Assert;
use CloudObjects\SDK\ObjectRetriever;
use ProgrammaBill\FilterQuerySet;
use ProgrammaBill\Entities\Transaction,
    ProgrammaBill\Entities\Actor,
    ProgrammaBill\Entities\Invoice,
    ProgrammaBill\Entities\PaymentInstruction;
use ProgrammaBill\AuthenticationSubscriber;    
use ProgrammaBill\InvoiceFormatter;
use ProgrammaBill\Services\ActorCreator;

/**
 * @Route("/api")
 */
class Transactions extends AbstractController {

    private $as;
    private $formatter;

    public function __construct(AuthenticationSubscriber $authenticationSubscriber,
            InvoiceFormatter $invoiceFormatter) {
        
        $this->as = $authenticationSubscriber;
        $this->formatter = $invoiceFormatter;
    }

    private function fetchTransaction($id) {
        // Fetch from database
        $transaction = $this->getDoctrine()
            ->getRepository(Transaction::class)
            ->findOneById($id);

        // Check whether entity exists and assert access rights
        if (!isset($transaction) || ($transaction->getMarketplace()
                != $this->as->getMarketplace()))
            throw $this->createNotFoundException();

        return $transaction;
    }

    /**
     * @Route("/transactions", methods="POST")
     */
    public function post(Request $r, ObjectRetriever $retriever, ActorCreator $actorCreator) {
        // Read and validate input
        $data = json_decode($r->getContent(), true);
        if (!isset($data) || !Transaction::isValidAPIRequest($data, [ '_actorless' => true ]))
            throw new BadRequestHttpException;

        // Create transaction
        $transaction = new Transaction($this->as->getMarketplace());
        $transaction->parseAPIRequest(null, $data, [
            '_actor_repository' => $this->getDoctrine()
                ->getRepository(Actor::class)
                ->setActorCreator($actorCreator),
            '_transaction_repository' => $this->getDoctrine()
                ->getRepository(Transaction::class),
            '_invoice_repository' => $this->getDoctrine()
                ->getRepository(Invoice::class),
            '_payment_instruction_repository' => $this->getDoctrine()
                ->getRepository(PaymentInstruction::class)
                ->setObjectRetriever($retriever)
        ]);

        if (!$r->headers->has('Execution') || $r->headers->get('Execution') != 'dry-run') {
            // Persist to database
            $this->getDoctrine()->getManager()->persist($transaction);
            $this->getDoctrine()->getManager()->flush();

            // Store invoices as PDF files
            foreach ($transaction->getInvoices() as $i)
                $this->formatter->createAndUploadPDF($i);

            return $this->json($transaction->createAPIResponse(null, [
                'expand_linked_transactions' => true,
                '_invoice_formatter' => $this->formatter                
            ]));
        } else {
            // This is a dry-run, so create response with no formatter = no PDF URLs
            return $this->json($transaction->createAPIResponse(null, [
                'expand_linked_transactions' => true
            ]));
        }
    }

    /**
     * @Route("/transactions", methods="GET")
     */
    public function getAll(Request $r) {
        $repository = $this->getDoctrine()->getRepository(Transaction::class);
               
        // Query repository
        $transactions = $repository->getAllForMarketplace(
            $this->as->getMarketplace(),
            FilterQuerySet::fromRequest($r, [ 'from', 'to' ])
        );
        
        // Iterate through transactions
        $items = [];
        foreach ($transactions as $t) {
            $items[] = $t->createAPIResponse(null, [
                'list_invoices_for_actor_only' => false,
                'expand_linked_transactions' => false,                
                '_invoice_formatter' => $this->formatter
            ]);
        }

        // Generate response
        return $this->json([
            'items' => $items
        ]);
    }

    /**
     * @Route("/transactions/{id}", methods="GET")
     */
    public function getOne($id) {
        // Fetch transaction
        $transaction = $this->fetchTransaction($id);

        // Generate response
        return $this->json($transaction->createAPIResponse(null, [
            'list_invoices_for_actor_only' => false,
            'embed_items' => true,
            'embed_links' => true,
            'expand_linked_transactions' => true,
            '_invoice_formatter' => $this->formatter
        ]));
    }

    /**
     * @Route("/transactions/{id}/actions", methods="POST")
     */
    public function processAction($id, Request $r) {
        // Fetch transaction
        $transaction = $this->fetchTransaction($id);

        // Prepare invoice repository
        $invoiceRepository = $this->getDoctrine()
            ->getRepository(Invoice::class);

        // Read and validate input
        $data = json_decode($r->getContent(), true);
        switch (@$data['action_type']) {
            case 'reverse':
                // Reverse the whole transaction

                // Reverse all invoices in transaction
                $reverseInvoices = [];
                foreach ($transaction->getInvoices() as $i) {
                    $invoice = $i->reverse($invoiceRepository);
                    Assert::object($invoice, "Could not reverse transaction, it may already be reversed.");
                    $reverseInvoices[] = $invoice;
                }
                
                foreach ($reverseInvoices as $i)
                    $this->formatter->createAndUploadPDF($i);

                break;
            case 'partial_reverse':
                // Partially reverse the transaction
                Assert::keyExists($data, 'types',
                    "For a partial reverse, use the 'types' parameter to specify the types to reverse.");

                // Reverse specified invoices in transaction
                $reverseInvoices = [];
                foreach ($transaction->getInvoices() as $i) {
                    if (in_array($i->getInvoiceTypeAsString(), $data['types'])) {
                        $invoice = $i->reverse($invoiceRepository);
                        Assert::object($invoice, "Could not reverse transaction, it may already be reversed.");
                        $reverseInvoices[] = $invoice;
                    }
                }

                Assert::minCount($reverseInvoices, "No invoices with given types found to reverse.");
            
                foreach ($reverseInvoices as $i)
                    $this->formatter->createAndUploadPDF($i);

                break;
            default:
                throw new BadRequestHttpException;
        }

        $this->getDoctrine()->getManager()->flush();

        // Generate response
        return $this->json($transaction->createAPIResponse(null, [
            'list_invoices_for_actor_only' => false,
            'embed_items' => true,
            'expand_linked_transactions' => true,
            '_invoice_formatter' => $this->formatter
        ]));
    }

}