<?php

namespace ProgrammaBill\MessageHandlers;

use Exception;
use Symfony\Component\Messenger\Handler\MessageHandlerInterface;
use Symfony\Component\Filesystem\Filesystem;
use Symfony\Component\Filesystem\Path;
use PhpZip\ZipFile;
use Doctrine\ORM\EntityManagerInterface;
use Webmozart\Assert\Assert;
use GuzzleHttp\Client;
use ProgrammaBill\InvoiceFormatter;
use ProgrammaBill\Messages\BulkExport;
use ProgrammaBill\Entities\Marketplace,
    ProgrammaBill\Entities\Actor,
    ProgrammaBill\Entities\Invoice;

class BulkExportHandler implements MessageHandlerInterface {

    private $em;
    private $formatter;
    private $client;

    public function __construct(EntityManagerInterface $em, InvoiceFormatter $formatter) {
        $this->em = $em;
        $this->formatter = $formatter;
        
        $this->client = new Client;
    }

    private function getInvoiceData(Invoice $invoice) {
        return $this->client
            ->get($this->formatter->getSignedPDFURL($invoice))
            ->getBody();
    }

    public function __invoke(BulkExport $export) {
        // Find entity to export
        $repository = $this->em->getRepository($export->getTargetType());
        Assert::notNull($repository);
        $entity = $repository->findOneById($export->getTargetId());
        Assert::notNull($entity);

        // Set options
        $options = array_merge([
            'outbound_folder' => 'Outbound',
            'inbound_folder' => 'Inbound'
        ], $export->getOptions());

        // Create a temporary directory to store exports
        $temporaryPath = sys_get_temp_dir() . '/' . date('YmdHi') . '_' . random_int(1000, 9999);
        $outboundPath = $temporaryPath . '/' . $options['outbound_folder'];
        $inboundPath = $temporaryPath . '/' . $options['inbound_folder'];
        
        $filesystem = new Filesystem;

        // Identify transactions
        if (is_a($entity, Marketplace::class)) {
            $query = $this->em->createQuery('SELECT t from ProgrammaBill\Entities\Transaction t WHERE t.marketplace = :marketplace AND t.created_timestamp >= :start_date AND t.created_timestamp <= :end_date');
            $query->setParameters([
                'marketplace' => $entity,
                'start_date' => $export->getStartDate(),
                'end_date' => $export->getEndDate()
            ]);
            $typeMap = [
                Invoice::TYPE_PLATFORM_TO_CUSTOMER => $outboundPath,
			    Invoice::TYPE_PLATFORM_TO_CUSTOMER_REVERSED => $outboundPath,
			    Invoice::TYPE_PLATFORM_TO_MERCHANT => $outboundPath,
			    Invoice::TYPE_PLATFORM_TO_MERCHANT_REVERSED => $outboundPath,
                Invoice::TYPE_MERCHANT_TO_PLATFORM => $inboundPath,
			    Invoice::TYPE_MERCHANT_TO_PLATFORM_REVERSED => $inboundPath
            ];
        } else {
            throw new Exception("Unsupported entity.");
        }

        // Process transactions and get invoice files
        $transactions = $query->getResult();

        foreach ($transactions as $t) {
            $invoices = $t->getInvoices();
            foreach ($invoices as $i) {
                if (isset($typeMap[$i->getInvoiceType()])) {
                    $filesystem->dumpFile(
                        Path::normalize($typeMap[$i->getInvoiceType()] . '/' . $i->getNumber() . '.pdf'),
                        $this->getInvoiceData($i)
                    );
                }
            }
        }

        // Package all files into a ZIP file
        $outputFile = Path::normalize($temporaryPath . '/' . 'Export_' . date('YmdHi') . '.zip');
        $archive = new ZipFile;
        $archive->addDirRecursive(Path::normalize($temporaryPath))
            ->saveAsFile($outputFile);

        // TODO: clean up; use try/catch/finally
        // TODO: put that file somewhere
        
        echo $outputFile . "\n";
        
    }
}