<?php

namespace ProgrammaBill\ProcessingAPI;

use Symfony\Bundle\FrameworkBundle\Controller\AbstractController;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Component\HttpFoundation\Request,
    Symfony\Component\HttpFoundation\Response;
use CloudObjects\SDK\ObjectRetriever;
use ProgrammaBill\Entities\Transaction,
    ProgrammaBill\Entities\Actor,
    ProgrammaBill\Entities\Invoice,
    ProgrammaBill\Entities\PaymentInstruction;
use ProgrammaBill\AuthenticationSubscriber;    
use ProgrammaBill\InvoiceFormatter;
use ProgrammaBill\Services\ActorCreator;

/**
 * @Route("/procapi")
 */
class Transactions extends AbstractController {

    private $as;
    private $formatter;

    public function __construct(AuthenticationSubscriber $authenticationSubscriber,
            InvoiceFormatter $invoiceFormatter) {
        
        $this->as = $authenticationSubscriber;
        $this->formatter = $invoiceFormatter;
    }

    private function fetchTransaction($id) {
        // Fetch from database
        $transaction = $this->getDoctrine()
            ->getRepository(Transaction::class)
            ->findOneById($id);

        // Check whether entity exists and assert access rights
        if (!isset($transaction) || ($transaction->getMarketplace()
                != $this->as->getActor()->getMarketplace()))
            throw $this->createNotFoundException();

        return $transaction;
    }

    /**
     * @Route("/transactions", methods="POST")
     */
    public function post(Request $r, ObjectRetriever $retriever, ActorCreator $actorCreator) {
        $actor = $this->as->getActor();

        // Read and validate input
        $data = json_decode($r->getContent(), true);
        if (!isset($data) || !Transaction::isValidAPIRequest($data))
            throw new BadRequestHttpException;

        // Create transaction
        $transaction = new Transaction($actor->getMarketplace());
        $transaction->parseAPIRequest($actor, $data, [
            '_actor_repository' => $this->getDoctrine()
                ->getRepository(Actor::class)
                ->setActorCreator($actorCreator),
            '_transaction_repository' => $this->getDoctrine()
                ->getRepository(Transaction::class),
            '_invoice_repository' => $this->getDoctrine()
                ->getRepository(Invoice::class),
            '_payment_instruction_repository' => $this->getDoctrine()
                ->getRepository(PaymentInstruction::class)
                ->setObjectRetriever($retriever)
        ]);

        // Persist to database
        $this->getDoctrine()->getManager()->persist($transaction);
        $this->getDoctrine()->getManager()->flush();

        // Store invoices as PDF files
        foreach ($transaction->getInvoices() as $i)
            $this->formatter->createAndUploadPDF($i, $actor);
    
        return $this->json($transaction->createAPIResponse($actor, [
            'list_invoices_for_actor_only' => false,
            'expand_linked_transactions' => true,
            '_invoice_formatter' => $this->formatter
        ]));
    }

    /**
     * @Route("/transactions/{id}", methods="GET")
     */
    public function getOne($id) {
        // Fetch transaction
        $transaction = $this->fetchTransaction($id);

        // Generate response
        return $this->json($transaction->createAPIResponse($this->as->getActor(), [
            'list_invoices_for_actor_only' => false,
            'expand_linked_transactions' => true,
            '_invoice_formatter' => $this->formatter
        ]));
    }

    /**
     * @Route("/transactions/{id}", methods="PUT")
     */
    public function put($id, Request $r) {
        // TODO: review this to check what is even allowed to be updated
        $actor = $this->as->getActor();

        // Read and validate input
        $data = json_decode($r->getContent(), true);
        if (!isset($data) || !Transaction::isValidAPIRequest($data))
            throw new BadRequestHttpException;

        // Fetch transaction
        $transaction = $this->fetchTransaction($id);
        $transaction->parseAPIRequest($actor, $data, [
            '_actor_repository' => $this->getDoctrine()
                ->getRepository(Actor::class),
            '_invoice_repository' => $this->getDoctrine()
                ->getRepository(Invoice::class)
        ]);

        // Persist to database
        $this->getDoctrine()->flush();

        // Store invoices as PDF files
        foreach ($transaction->getInvoices() as $i)
            $this->formatter->createAndUploadPDF($i, $actor); // TODO: as an optimization, only update invoices that have been changed

        return $this->json($transaction->createAPIResponse($actor, [
            'list_invoices_for_actor_only' => false,
            '_invoice_formatter' => $this->formatter
        ]));
    }

}
