<?php

namespace ProgrammaBill\Repositories;

use Doctrine\ORM\EntityRepository, Doctrine\ORM\NoResultException;
use CloudObjects\SDK\AccountGateway\AAUIDParser, CloudObjects\SDK\AccountGateway\AccountContext;
use ProgrammaBill\Entities\Actor, ProgrammaBill\Entities\Marketplace;
use ProgrammaBill\Services\ActorCreator;

class ActorRepository extends EntityRepository {

	private $actorCreator;

	public function linkWithExistingAccounts(Actor $actor) {
		// Check if the marketplace wants auto-linking this account
		$marketplaceRepository = $this->_em->getRepository(Marketplace::class);		
		
		$marketplace = $actor->getMarketplace();
		$marketplaceCoid = $marketplace->getCOID();
		$targetMarketplaces = $marketplace->getAutolinkServiceCOIDsForMatchingEMail();
		
		foreach ($targetMarketplaces as $targetCoid) {
			if ($targetCoid->equals($marketplaceCoid)) {
				// ignore circular link
				continue;
			}
			$targetMarketplace = $marketplaceRepository->fetchOrCreateWithCOID($targetCoid)
				->setObjectRetriever($marketplace->getObjectRetriever());
			// Only apply when linking is bidirectional
			foreach ($targetMarketplace->getAutolinkServiceCOIDsForMatchingEMail() as $backlinkCoid) {
				if ($backlinkCoid->equals($marketplaceCoid)) {
					// Mapping established, now let's see if there are accounts
					$targetActor = $this->fetchWithEmail($actor->getEMail(), $targetMarketplace);
					if (isset($targetActor)) {
						// Found target actor
						$identity = $targetActor->getOrCreateIdentity();
						if ($identity->getId() == null) {
							$this->_em->persist($identity);
						}
						$actor->setIdentity($identity);
						$identity->getActors()->add($actor);
					}
				}
			}
		}
	}

	public function setActorCreator(ActorCreator $actorCreator) {
		$this->actorCreator = $actorCreator;
		return $this;
	}

	public function fetchWithEmail(string $email, Marketplace $marketplace) {
		try {
			$query = $this->_em->createQuery('SELECT a FROM ProgrammaBill\Entities\Actor a WHERE a.email = :email AND a.marketplace = :marketplace');
			$query->setParameter('email', $email)
				->setParameter('marketplace', $marketplace);
			return $query->getSingleResult();
		} catch (NoResultException $e) {
			return null;
		}
	}

	public function fetchOrCreateWithEmail(string $email, Marketplace $marketplace) {
		$actor = $this->fetchWithEmail($email, $marketplace);
		
		if (!isset($actor)) {
			$actor = $this->actorCreator->createActorForMarketplaceWithEMail($marketplace, $email);
			$this->linkWithExistingAccounts($actor);
			$this->_em->persist($actor);
		}

		return $actor;
	}

	public function fetchOrCreateWithExternalUserID(string $ext_user_id, Marketplace $marketplace, string $emailFallback = null) {
		try {
			$query = $this->_em->createQuery('SELECT a FROM ProgrammaBill\Entities\Actor a WHERE a.ext_user_id = :ext_user_id AND a.marketplace = :marketplace');
			$query->setParameter('ext_user_id', $ext_user_id)
				->setParameter('marketplace', $marketplace);
			return $query->getSingleResult();
		} catch (NoResultException $e) {
			if (isset($emailFallback)) {
				// We have a fallback, so try again with that before creating
				$actor = $this->fetchWithEmail($emailFallback, $marketplace);
				// If found, add external user ID to actor
				if (isset($actor)) {
					$actor->setExternalUserID($ext_user_id);
					return $actor;
				}					
			}
			$actor = $this->actorCreator->createActorForMarketplaceWithExternalUserID($marketplace, $ext_user_id);
			if (isset($emailFallback)) {
				$actor->setEMail($emailFallback); // must set this so we can check for links, other fields set later
			}
			$this->linkWithExistingAccounts($actor);
			$this->_em->persist($actor);
			return $actor;
		}
	}

	public function fetchOrCreateFor(AccountContext $context) {
    	// TODO: detect and support group requests
		try {
			$aauid = AAUIDParser::getAAUID($context->getAAUID());
			$query = $this->_em->createQuery('SELECT a FROM ProgrammaBill\Entities\Actor a WHERE a.aauid = :aauid');
			$query->setParameter('aauid', $aauid);

			return $query->getSingleResult();
		} catch (NoResultException $e) {
			// Account doesn't exist, create ...
			$marketplaceDomain = substr($context->getConnectedAccount()
				->getProperty('coid://aauid.net/isForService')->getId(), 7);
			$marketplace = $this->_em->getRepository('ProgrammaBill\Entities\Marketplace')
				->fetchOrCreateWithNamespace($marketplaceDomain);

			$actor = new Actor($marketplace, $aauid, Actor::TYPE_INDIVIDUAL);
			$actor->setAccessToken($context->getAccessToken());
			$this->_em->persist($actor);
			return $actor;
		}
	}

}
