<?php

namespace ProgrammaBill\Repositories;

use DateTime, Exception;
use Carbon\Carbon;
use Doctrine\ORM\EntityRepository, Doctrine\ORM\NoResultException;
use Doctrine\Common\Collections\ArrayCollection;
use Webmozart\Assert\Assert;
use ProgrammaBill\Entities\Actor, ProgrammaBill\Entities\Transaction,
	ProgrammaBill\Entities\Marketplace, ProgrammaBill\Entities\Invoice,
	ProgrammaBill\Entities\InvoiceAction;

class InvoiceRepository extends EntityRepository {

	private function getAllMerchantTypes() {
		return [
			Invoice::TYPE_MERCHANT_TO_CUSTOMER,
			Invoice::TYPE_MERCHANT_TO_CUSTOMER_REVERSED,
			Invoice::TYPE_MERCHANT_TO_PLATFORM,
			Invoice::TYPE_MERCHANT_TO_PLATFORM_REVERSED
		];
	}

	private function getAllPlatformTypes() {
		return [
			Invoice::TYPE_PLATFORM_TO_CUSTOMER,
			Invoice::TYPE_PLATFORM_TO_CUSTOMER_REVERSED,
			Invoice::TYPE_PLATFORM_TO_MERCHANT,
			Invoice::TYPE_PLATFORM_TO_MERCHANT_REVERSED
		];
	}

	/**
	 * Counts the number of invoices that a merchant has created in a certain interval.
	 * @param Actor $merchant The merchant.
	 * @param DateTime $start Start of the interval.
	 * @param DateTime $end End of the interval.
	 * @return integer
	 */
	public function countInvoicesForMerchant(Actor $merchant, DateTime $start, DateTime $end) {
		$actors = ($merchant->getIdentity() == null)
			? [ $merchant ] // singleton account without identity
			: $merchant->getIdentity()->getActors(); // all actors in identity are counted

		// TODO: depending on invoicing rules, not all accounts may be required to query
		$actorsForQuery = [];
		foreach ($actors as $a) {
			if ($a->getId() != null) {
				$actorsForQuery[] = $a;
			}
		}
	
		if (count($actorsForQuery) == 0) {
			// no actors we can query
			return 0;
		}		

		$query = $this->_em->createQuery("SELECT COUNT(i) FROM ProgrammaBill\Entities\Invoice i JOIN i.transaction t "
			."WHERE i.invoice_type IN (:type) AND t.merchant IN (:actors) AND i.created_timestamp > :start AND i.created_timestamp < :end");
		$query->setParameters([
			'type' => $this->getAllMerchantTypes(),
			'actors' => $actorsForQuery,
			'start' => $start,
			'end' => $end
		]);

		return $query->getSingleScalarResult();
	}

	/**
	 * Retrieves all invoices from a merchant with an event in a certain interval.
	 * @param Actor $merchant The merchant.
	 * @param string $event One of: "created", "payment_received", "paid_out"
	 * @param DateTime $start Start of the interval.
	 * @param DateTime $end End of the interval.
	 */
	public function getInvoicesForMerchant(Actor $merchant, string $event,
			DateTime $start, DateTime $end) {
		
		if ($event == 'created') {
			$query = $this->_em->createQuery("SELECT i FROM ProgrammaBill\Entities\Invoice i JOIN i.transaction t "
				."WHERE i.invoice_type IN (:type) AND t.merchant = :actor AND i.created_timestamp > :start AND i.created_timestamp < :end");
			$query->setParameters([
				'type' => $this->getAllMerchantTypes(),
				'actor' => $merchant,
				'start' => $start,
				'end' => $end
			]);
		} elseif ($event == 'payment_received' || $event == 'paid_out') {
			$query = $this->_em->createQuery("SELECT i FROM ProgrammaBill\Entities\Invoice i JOIN i.transaction t JOIN i.actions a "
				."WHERE i.invoice_type IN (:type) AND t.merchant = :actor AND a.created_timestamp > :start AND a.created_timestamp < :end AND a.action_type IN (:actions)");
			
			$actions = ($event == 'payment_received')
				? [ InvoiceAction::TYPE_PAYIN, InvoiceAction::TYPE_PAYDIRECT ]
				: [ InvoiceAction::TYPE_PAYOUT, InvoiceAction::TYPE_PAYDIRECT ];
			$query->setParameters([
				'type' => $this->getAllMerchantTypes(),
				'actor' => $merchant,
				'start' => $start,
				'end' => $end,
				'actions' => $actions
			]);
		} else
			throw new Exception("Invalid 'event' parameter.");

		return $query->getResult();
	}

	/**
	 * Counts the number of platform invoices that a marketplace has created in a certain interval.
	 * @param Marketplace $marketplace The marketplace
	 * @param DateTime $start Start of the interval.
	 * @param DateTime $end End of the interval.
	 * @return integer
	 */
	public function countInvoicesForMarketplace(Marketplace $marketplace, DateTime $start, DateTime $end) {
		$query = $this->_em->createQuery("SELECT COUNT(i) FROM ProgrammaBill\Entities\Invoice i JOIN i.transaction t "
			."WHERE i.invoice_type IN (:type) AND t.marketplace = :marketplace AND i.created_timestamp > :start AND i.created_timestamp < :end");
		$query->setParameters([
			'type' => $this->getAllPlatformTypes(),
			'marketplace' => $marketplace,
			'start' => $start,
			'end' => $end
		]);
		return $query->getSingleScalarResult();
	}

	private function generateNumber($cycle, Transaction $transaction, $invoice_type) {
		switch ($cycle) {
			case "d":
				// Daily
				$start = Carbon::now()->startOfDay();
				$end = Carbon::now()->endOfDay();
				$prefix = date('Ymd');
				break;
			case "m":
				// Monthly
				$start = Carbon::now()->startOfMonth();
				$end = Carbon::now()->endOfMonth();
				$prefix = date('Ym');
				break;
			case "Y":
				// Yearly
				$start = Carbon::now()->startOfYear();
				$end = Carbon::now()->endOfYear();
				$prefix = date('Y');
				break;
		}
		if (in_array($invoice_type, $this->getAllMerchantTypes())) {
			// Count invoices from merchant to determine next number
			$number = $this->countInvoicesForMerchant($transaction->getMerchant(), $start, $end) + 1;
		} else {
			// Count invoices from marketplaces as well as invoices already
			// created for this transaction to determine next number

			$firstNumber = $transaction->getMarketplace()->getFirstInvoiceNumber();
			if (isset($firstNumber)) {
				// This marketplace has an offset defined; check if there are invoices
				// in previous segments
				$previousTransactions = $this->countInvoicesForMarketplace(
					$transaction->getMarketplace(), (new DateTime('2020-01-01')), $start);
				
				if ($previousTransactions == 0) {
					// If none (= this is the first year/month) the marketplace is active:
					// apply offset
					$number = $firstNumber;
				} else
					$number = 1;
			} else
				$number = 1;

			$number += $this->countInvoicesForMarketplace($transaction->getMarketplace(),
				$start, $end);

			foreach ($transaction->getInvoices() as $i) {
				if (!$i->getId() && in_array($i->getInvoiceType(), $this->getAllPlatformTypes()))
					$number++;
			}
		}
		
		return $prefix."-".$number;
	}

	public function createNewInvoiceForMerchant(Transaction $transaction, $invoice_type) {
		Assert::inArray($invoice_type, $this->getAllMerchantTypes());

		return new Invoice($transaction, $invoice_type,
			$this->generateNumber($transaction->getMerchant()->getInvoicingCycle(),
				$transaction, $invoice_type));
	}

	public function createNewInvoiceForMarketplace(Transaction $transaction, $invoice_type) {
		Assert::inArray($invoice_type, $this->getAllPlatformTypes());
		
		return new Invoice($transaction, $invoice_type,
			$this->generateNumber($transaction->getMarketplace()->getInvoicingCycle(),
				$transaction, $invoice_type));
	}

	public function getM2CWithMarketplaceMerchantEMailAndNumber(Marketplace $marketplace,
			$merchant_email, $number) {
		$query = $this->_em->createQuery("SELECT i FROM ProgrammaBill\Entities\Invoice i, ProgrammaBill\Entities\Transaction t "
            ."JOIN t.merchant m WHERE i.transaction = t AND i.number = :number AND t.marketplace = :marketplace "
            ."AND m.email = :merchant_email AND i.invoice_type IN (:type)");        
        $query->setParameters([
            'number' => $number,
            'marketplace' => $marketplace,
            'merchant_email' => $merchant_email,
            'type_m2c' => $this->getAllMerchantTypes()
        ]);

        try {
            return $query->getSingleResult();
        } catch (NoResultException $e) {
            return null;
        }
	}

	public function getPlatformInvoiceWithNumber(Marketplace $marketplace, $number) {
		$query = $this->_em->createQuery("SELECT i FROM ProgrammaBill\Entities\Invoice i JOIN i.transaction t "
            ."WHERE i.invoice_type IN (:type) AND t.marketplace = :marketplace AND i.number = :number");
        $query->setParameters([
            'number' => $number,
            'marketplace' => $marketplace,
            'type' => $this->getAllPlatformTypes()
        ]);

        try {
            return $query->getSingleResult();
        } catch (NoResultException $e) {
            return null;
        }
	}

	/*public function getAllOverdueInvoices(Marketplace $marketplace) {
		$query = $this->_em->createQuery("SELECT i FROM ProgrammaBill\Entities\Invoice i JOIN i.transaction t "
			."WHERE i.due_date < :date AND i.is_payment_received = :paid AND t.marketplace = :marketplace");
		$query->setParameters([
			'date' => Carbon::now()->startOfDay()->subDays($marketplace->getOverdueDelay()),
			'marketplace' => $marketplace,
			'paid' => false
		]);	

		return new ArrayCollection($query->getResult());
	} */

	public function getAllOverdueInvoices(Marketplace $marketplace) {
		$query = $this->_em->createQuery("SELECT i FROM ProgrammaBill\Entities\Invoice i JOIN i.transaction t "
			."WHERE i.due_date < :date AND i.is_payment_received = :paid AND t.marketplace = :marketplace "
			."AND i.created_timestamp < :end AND i.created_timestamp > :start");
		$query->setParameters([
			'date' => Carbon::now()->startOfDay()->subDays($marketplace->getOverdueDelay()),
			'end' => Carbon::now()->subMonth(1)->startOfMonth(),
			'start' => Carbon::parse('2023-07-04'),
			'marketplace' => $marketplace,
			'paid' => false
		]);	

		$collection = new ArrayCollection($query->getResult());

		// Before returning, filter reversed invoices because they don't count as overdue
		return $collection->filter(function($invoice) {
			return (count($invoice->findActionsWithType(InvoiceAction::TYPE_REVERSE)) == 0);
		});
	}

}
