<?php

namespace ProgrammaBill;

use Exception;
use Webmozart\Assert\Assert;
use Punic\Territory;
use ProgrammaBill\Interfaces\Address;

/**
 * The TemplateFinder encapsulates the rules used for selecting a specific template.
 */
class TemplateFinder {

    const INVOICE = 'invoice';

    /**
     * Evaluates sender and receiver and returns the filename of
     * the template that should be used.
     * 
     * @param string $purpose The type of the document.
     * @param Address $sender Sender of the document.
     * @param Address $receiver Receiver of the document.
     * @param array $extra Additional information
     */
    public static function getTemplateFilenameFor(string $purpose, Address $sender,
            Address $receiver, array $extra = []) {

        Assert::oneOf($purpose, [ self::INVOICE ]);
        
        $allCountries = array_keys(Territory::getList('c'));
        $fromCountry = $sender->getCountry();
        $toCountry = $receiver->getCountry();

        Assert::oneOf($fromCountry, $allCountries);
        Assert::oneOf($toCountry, $allCountries);

        switch ($fromCountry) {
            case "DE": // Germany
            case "AT": // Austria
                // These countries have one template for domestic
                // (local language) and one for international
                // (local language + target or English) invoicing
                
                // Also, there are different templates for
                // business and individual customers
                $suffix = isset($extra['suffix']) ? $extra['suffix']
                    : ($receiver->isBusiness() ? 'b2b' : 'b2c');

                if ($fromCountry == $toCountry) {
                    // Local transaction
                    return "default/".$fromCountry."/invoice-domestic-".$suffix.".html.twig";
                } elseif (in_array($fromCountry, [ 'DE', 'AT' ]) && in_array($toCountry, [ 'DE', 'AT' ])) {
                    // AT and DE use the same language (German), so we can use the local template
                    return "default/".$fromCountry."/invoice-domestic-".$suffix.".html.twig";
                } else {
                    // International transaction
                    return "default/".$fromCountry."/invoice-international-".$suffix.".html.twig";
                }
            default:
                // Other countries
                throw new Exception("not yet supported!");
        }
    }
}