<?php

namespace ProgrammaBill\Entities;

use DateTime;
use PHPUnit\Framework\TestCase;

class InvoiceTest extends TestCase {

    private $transaction;

    public function setUp() : void {
        $marketplace = new Marketplace('example.com');
        $this->transaction = new Transaction($marketplace);
        
        $merchant = new Actor($marketplace, '1234123412341234', Actor::TYPE_UNKNOWN);
        $this->transaction->setMerchant($merchant);
    }

    public function testTaxLessInvoiceCalculation() : void {
        $invoice = new Invoice($this->transaction, Invoice::TYPE_MERCHANT_TO_CUSTOMER, 'Test');
        $invoice->setDueDate(new DateTime);
        $invoice->addItem('Sample Good A', 50, 1, '', 0);
        $invoice->addItem('Sample Good B', 100, 1, '', 0);

        $this->assertEquals(150, $invoice->getTotalPayableAmount());
        $this->assertEquals(150, $invoice->getTotalPretaxAmount());
        $this->assertEquals(150, $invoice->getPayableAmountForRate(0));
        $this->assertEquals(0, $invoice->getTotalTaxAmount());        
        $this->assertEquals([ 0 ], $invoice->getTaxRates());

        // 0 != 'R'
        $this->assertEquals(0, $invoice->getPayableAmountForRate('R'));

        $response = $invoice->createAPIResponse(null, [
            'embed_items' => true,
            'embed_tax_summary' => true            
        ]);

        $item = $response['items'][0];
        $this->assertEquals(1, $item['count']);
        $this->assertEquals(0, $item['tax_rate']);
        $this->assertEquals(50, $item['unit_price_without_tax']);
        $this->assertEquals(50, $item['unit_price_with_tax']);
        $this->assertEquals(50, $item['total_price']);
        $this->assertEquals(50, $item['total_price_without_tax']);
        $this->assertEquals(50, $item['total_price_with_tax']);

        $summary = $response['tax_summary'][0];
        $this->assertEquals(0, $summary['rate']);
        $this->assertEquals(0, $summary['tax_amount']);
        $this->assertEquals(150, $summary['pretax_amount']);
        $this->assertEquals(150, $summary['total_amount']);
    }

    public function testReverseChargeInvoiceCalculation() : void {
        $invoice = new Invoice($this->transaction, Invoice::TYPE_MERCHANT_TO_CUSTOMER, 'Test');
        $invoice->setDueDate(new DateTime);
        $invoice->addItem('Sample Good A', 50, 1, '', 'R');
        $invoice->addItem('Sample Good B', 100, 1, '', 'R');

        $this->assertEquals(150, $invoice->getTotalPayableAmount());
        $this->assertEquals(150, $invoice->getTotalPretaxAmount());
        $this->assertEquals(150, $invoice->getPayableAmountForRate('R'));
        $this->assertEquals(0, $invoice->getTotalTaxAmount());        
        $this->assertEquals([ 'R' ], $invoice->getTaxRates());

        // 0 != 'R'
        $this->assertEquals(0, $invoice->getPayableAmountForRate(0));

        $response = $invoice->createAPIResponse(null, [
            'embed_items' => true,
            'embed_tax_summary' => true            
        ]);

        $item = $response['items'][0];
        $this->assertEquals(1, $item['count']);
        $this->assertEquals('R', $item['tax_rate']);
        $this->assertEquals(50, $item['unit_price_without_tax']);
        $this->assertEquals(50, $item['unit_price_with_tax']);
        $this->assertEquals(50, $item['total_price']);
        $this->assertEquals(50, $item['total_price_without_tax']);
        $this->assertEquals(50, $item['total_price_with_tax']);

        $summary = $response['tax_summary'][0];
        $this->assertEquals('R', $summary['rate']);
        $this->assertEquals(0, $summary['tax_amount']);
        $this->assertEquals(150, $summary['pretax_amount']);
        $this->assertEquals(150, $summary['total_amount']);
    }

    public function testSingleRateCalculationTaxExcluded() : void {
        $invoice = new Invoice($this->transaction, Invoice::TYPE_MERCHANT_TO_CUSTOMER, 'Test', false);
        $invoice->setDueDate(new DateTime);
        $invoice->addItem('Sample Good A', 50, 1, '', 10);
        $invoice->addItem('Sample Good B', 100, 1, '', 10);

        $this->assertEquals(165, $invoice->getTotalPayableAmount());
        $this->assertEquals(150, $invoice->getTotalPretaxAmount());
        $this->assertEquals(165, $invoice->getPayableAmountForRate(10));
        $this->assertEquals(15, $invoice->getTotalTaxAmount());        
        $this->assertEquals([ 10 ], $invoice->getTaxRates());

        $response = $invoice->createAPIResponse(null, [
            'embed_items' => true,
            'embed_tax_summary' => true            
        ]);

        $item = $response['items'][0];
        $this->assertEquals(1, $item['count']);
        $this->assertEquals(10, $item['tax_rate']);
        $this->assertEquals(50, $item['unit_price_without_tax']);
        $this->assertEquals(55, $item['unit_price_with_tax']);
        $this->assertEquals(50, $item['total_price']);
        $this->assertEquals(50, $item['total_price_without_tax']);
        $this->assertEquals(55, $item['total_price_with_tax']);

        $summary = $response['tax_summary'][0];
        $this->assertEquals(10, $summary['rate']);
        $this->assertEquals(15, $summary['tax_amount']);
        $this->assertEquals(150, $summary['pretax_amount']);
        $this->assertEquals(165, $summary['total_amount']);
    }

    public function testSingleRateCalculationTaxIncluded() : void {
        $invoice = new Invoice($this->transaction, Invoice::TYPE_MERCHANT_TO_CUSTOMER, 'Test', true);
        $invoice->setDueDate(new DateTime);
        $invoice->addItem('Sample Good A', 50, 1, '', 10);
        $invoice->addItem('Sample Good B', 100, 1, '', 10);

        $this->assertEquals(150, $invoice->getTotalPayableAmount());
        $this->assertEquals(136.36, $invoice->getTotalPretaxAmount());
        $this->assertEquals(150, $invoice->getPayableAmountForRate(10));
        $this->assertEquals(13.64, $invoice->getTotalTaxAmount());        
        $this->assertEquals([ 10 ], $invoice->getTaxRates());

        $response = $invoice->createAPIResponse(null, [
            'embed_items' => true,
            'embed_tax_summary' => true            
        ]);

        $item = $response['items'][0];
        $this->assertEquals(1, $item['count']);
        $this->assertEquals(10, $item['tax_rate']);
        $this->assertEquals(45.45, $item['unit_price_without_tax']);
        $this->assertEquals(50, $item['unit_price_with_tax']);
        $this->assertEquals(50, $item['total_price']);
        $this->assertEquals(45.45, $item['total_price_without_tax']);
        $this->assertEquals(50, $item['total_price_with_tax']);

        $summary = $response['tax_summary'][0];
        $this->assertEquals(10, $summary['rate']);
        $this->assertEquals(13.64, $summary['tax_amount']);
        $this->assertEquals(136.36, $summary['pretax_amount']);
        $this->assertEquals(150, $summary['total_amount']);
    }

    public function testSingleRateCalculationTaxIncludedMultiItem() : void {
        $invoice = new Invoice($this->transaction, Invoice::TYPE_MERCHANT_TO_CUSTOMER, 'Test', true);
        $invoice->setDueDate(new DateTime);
        $invoice->addItem('Sample Good A', 50, 3, '', 10);
        $invoice->addItem('Sample Good B', 100, 1, '', 10);

        $this->assertEquals(250, $invoice->getTotalPayableAmount());
        $this->assertEquals(227.27, $invoice->getTotalPretaxAmount());
        $this->assertEquals(250, $invoice->getPayableAmountForRate(10));
        $this->assertEquals(22.73, $invoice->getTotalTaxAmount());        
        $this->assertEquals([ 10 ], $invoice->getTaxRates());

        $response = $invoice->createAPIResponse(null, [
            'embed_items' => true,
            'embed_tax_summary' => true            
        ]);

        $item = $response['items'][0];
        $this->assertEquals(3, $item['count']);
        $this->assertEquals(10, $item['tax_rate']);
        $this->assertEquals(45.45, $item['unit_price_without_tax']);
        $this->assertEquals(50, $item['unit_price_with_tax']);
        $this->assertEquals(150, $item['total_price']);
        $this->assertEquals(136.35, $item['total_price_without_tax']);
        $this->assertEquals(150, $item['total_price_with_tax']);

        $summary = $response['tax_summary'][0];
        $this->assertEquals(10, $summary['rate']);
        $this->assertEquals(22.73, $summary['tax_amount']);
        $this->assertEquals(227.27, $summary['pretax_amount']);
        $this->assertEquals(250, $summary['total_amount']);
    }

}