<?php

namespace ProgrammaBill\Entities;

use DateTime;
use PHPUnit\Framework\TestCase;

class CreateTransactionRequestTest extends TestCase {

    const INVOICE_NUMBER = '0000';

    private $transaction;
    private $actor;
    private $actorRepository;
    private $invoiceRepository;

    private function getRequestContent($filename) {
        $content = file_get_contents(__DIR__ . '/json/'. $filename . '.json');
        return json_decode($content, true);
    }

    public function setUp() : void {
        $marketplace = new Marketplace('example.com');
        $this->transaction = new Transaction($marketplace);
        
        $this->actor = new Actor($marketplace, '1234123412341234', Actor::TYPE_UNKNOWN);

        // Mock repositories
        $this->actorRepository = $this->createMock('ProgrammaBill\Repositories\ActorRepository');
        $this->actorRepository->expects($this->any())
            ->method('fetchOrCreateWithExternalUserID')
            ->willReturn($this->actor);

        $this->invoiceRepository = $this->createMock('ProgrammaBill\Repositories\InvoiceRepository');
        $this->invoiceRepository->expects($this->any())
            ->method('createNewInvoiceForMarketplace')
            ->willReturn(new Invoice($this->transaction, Invoice::TYPE_PLATFORM_TO_CUSTOMER, self::INVOICE_NUMBER));
    }

    public function testCreateMinimalP2CTransaction() : void {
        $request = $this->getRequestContent('create_min_p2c_transaction');
        
        $this->assertTrue(Transaction::isValidAPIRequest($request, [
            '_actorless' => true
        ]));

        $t = $this->transaction;        

        $t->parseAPIRequest(null, $request, [
            '_actor_repository' => $this->actorRepository,
            '_invoice_repository' => $this->invoiceRepository
        ]);

        // Validate customer assignment
        $this->assertNull($t->getMerchant());
        $this->assertNotNull($t->getCustomer());
        $this->assertEquals($this->actor, $t->getCustomer());

        // Validate transfer of customer data
        $c = $t->getCustomer();
        $cData = $request['customer'];
        $this->assertEquals($cData['email'], $c->getEMail());
        $this->assertEquals($cData['org_name'], $c->getOrgName());
        $this->assertEquals($cData['full_name'], $c->getFullName());
        // ...

        // Validate creation of one (1) invoice
        $this->assertCount(1, $t->getInvoices());
        $this->assertNotNull($t->getInvoiceWithType(Invoice::TYPE_PLATFORM_TO_CUSTOMER));

        // Validate invoice details
        $i = $t->getInvoiceWithType(Invoice::TYPE_PLATFORM_TO_CUSTOMER);
        $iData = $request['invoices'][0];
        $this->assertEquals($iData['currency'], $i->getCurrency());
        $this->assertEquals(self::INVOICE_NUMBER, $i->getNumber());
        $this->assertCount(count($iData['items']), $i->getItems());
    }

}